# -*- coding: utf-8 -*-

# Keithley.py driver for Keithley 2636A multi channel source measure unit
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
# Micha Wildermuth, micha.wildermuth@kit.edu 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import numpy as np
import time
import logging
from distutils.version import LooseVersion


class Keithley(Instrument):
    """
    This is the driver for the Keithley 2636A multi channel source measure unit.
    """
    
    def __init__(self, name, address, reset=False):
        """
        Initializes VISA communication with the instrument Keithley 2636A.
        
        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        address: string
            GPIB address for the communication with the instrument.
        reset: bool, optional
            Resets the instrument to default conditions. Default is False.
        
        Returns
        -------
        None
        
        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        
        >>> IVD = qkit.instruments.create('IVD', 'Keithley', address='TCPIP0::00.00.000.00::INSTR', reset=True)
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        self.__name__ = __name__
        # Start VISA communication
        logging.info(__name__ + ': Initializing instrument Keithley_2636A')
        Instrument.__init__(self, name, tags=['physical'])
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        # distinguish high level commands to query status bit for different pyvisa versions
        if LooseVersion(visa.__version__) < LooseVersion('1.5.0'):  # pyvisa 1.4
            self._read_stb = lambda: visa.vpp43.read_stb(self._visainstrument.vi)
        else:  # pyvisa 1.5
            self._read_stb = lambda: visa.visalib.read_stb(self._visainstrument.session)[0]
        # internal variables
        self._sweep_mode = 1  # IV-mode
        self._sweep_channels = (1,)
        self._readingbuffer_bias = None
        self._readingbuffer_sense = None
        self._readingbuffer = ()
        self._measurement_modes = {0: '2-wire', 1: '4-wire'}
        self._mode_types = {0: 'curr', 1: 'volt', 2: 'res', 3: 'pow'}
        self._smu_command = {0: 'i', 1: 'v'}
        self._IV_units = {0: 'A', 1: 'V'}
        self._avg_types = ['moving average', 'repeat average', 'median']
        # dict of defaults values: defaults[<sweep_mode>][<channel>][<parameter>][<value>]
        self._defaults = {0: [{self.set_measurement_mode: 0,  # bias channel
                               self.set_bias_mode:  1,
                               self.set_sense_mode: 1,
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 15e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0},
                              {self.set_measurement_mode: 0,  # sense channel
                               self.set_bias_mode: 0,
                               self.set_sense_mode: 1,
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 15e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0}],
                          1: [{self.set_measurement_mode: 1,
                               self.set_bias_mode: 0,
                               self.set_sense_mode: 1,
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 15e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0}],
                          2: [{self.set_measurement_mode: 1,
                               self.set_bias_mode: 1,
                               self.set_sense_mode: 0,
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 15e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0}]}
        self._default_str = 'default parameters'
        self.__set_defaults_docstring()
        # error messages
        self.err_msg = {-430: '''Query Deadlocked''',
                        -420: '''Query Unterminated''',
                        -410: '''Query Interrupted''',
                        -363: '''Input Buffer Over-run''',
                        -350: '''Queue Overflow''',
                        -315: '''Configuration Memory Lost''',
                        -314: '''Save/ Recall Memory Lost''',
                        -292: '''Referenced name does not exist''',
                        -286: '''TSP Runtime error''',
                        -285: '''Program Syntax''',
                        -282: '''Illegal program name''',
                        -281: '''Cannot Create Program''',
                        -225: '''Out of Memory or TSP Memory allocation error''',
                        -224: '''Illegal Parameter Value''',
                        -223: '''Too Much Data''',
                        -222: '''Parameter Data Out of Range''',
                        -221: '''Settings Conflict''',
                        -220: '''Parameter''',
                        -203: '''Command protected''',
                        -154: '''String Too Long''',
                        -151: '''Invalid String Data''',
                        -144: '''Character Data Too Long''',
                        -141: '''Invalid Character Data''',
                        -121: '''Invalid Character In Number''',
                        -120: '''Numeric Data''',
                        -109: '''Missing Parameter''',
                        -108: '''Parameter Not Allowed''',
                        -105: '''Trigger Not Allowed''',
                        -104: '''Data Type''',
                        -101: '''Invalid Character''',
                        0: '''Queue Is Empty''',
                        603: '''Power On State Lost''',
                        702: '''Unresponsive digital FPGA''',
                        802: '''Output Blocked By Interlock''',
                        820: '''Parsing Value''',
                        900: '''Internal System''',
                        1100: '''Command Unavailable''',
                        1101: '''Parameter Too Big''',
                        1102: '''Parameter Too Small''',
                        1103: '''Max Greater Than Min''',
                        1104: '''Too many digits for param type''',
                        1106: '''Battery Not Present''',
                        1107: '''Cannot modify factory menu''',
                        1108: '''Menu name does not exist''',
                        1109: '''Menu name already exists''',
                        1110: '''Catastrophic analog supply failure''',
                        1200: '''TSPlink initialization failed''',
                        1201: '''TSPlink initialization failed''',
                        1202: '''TSPlink initialization failed''',
                        1203: '''TSPlink initialization failed (possible loop in node chain)''',
                        1204: '''TSPlink initialization failed''',
                        1205: '''TSPlink initialization failed (no remote nodes found)''',
                        1206: '''TSPlink initialization failed''',
                        1207: '''TSPlink initialization failed''',
                        1208: '''TSPlink initialization failed''',
                        1209: '''TSPlink initialization failed''',
                        1210: '''TSPlink initialization failed (node ID conflict)''',
                        1211: '''Node %u is inaccessible''',
                        1212: '''Invalid node ID''',
                        1400: '''Expected at least %d parameters''',
                        1401: '''Parameter %d is invalid''',
                        1402: '''User scripts lost''',
                        1403: '''Factory scripts lost''',
                        1404: '''Invalid byte order''',
                        1405: '''Invalid ASCII precision''',
                        1406: '''Invalid data format''',
                        1500: '''Invalid baud rate setting''',
                        1501: '''Invalid parity setting''',
                        1502: '''Invalid terminator setting''',
                        1503: '''Invalid bits setting''',
                        1504: '''Invalid flow control setting''',
                        1600: '''Maximum GPIB message length exceeded''',
                        1700: '''Display area boundary exceeded''',
                        1800: '''Invalid Digital Trigger Mode''',
                        1801: '''Invalid digital I/O Line''',
                        2000: '''Flash download error''',
                        2001: '''Cannot flash with error in queue''',
                        2100: '''Could not open socket''',
                        2101: '''Could not close socket''',
                        2102: '''LAN configuration already in progress''',
                        2103: '''LAN disabled''',
                        2104: '''Socket error''',
                        2105: '''Unreachable gateway''',
                        2106: '''Could not acquire ip address''',
                        2107: '''Duplicate IP address detected''',
                        2108: '''DHCP lease lost''',
                        2109: '''LAN cable disconnected''',
                        2110: '''Could not resolve hostname''',
                        2111: '''DNS name (FQDN) too long''',
                        2112: '''Connection not established''',
                        2200: '''File write error''',
                        2201: '''File read error''',
                        2202: '''Cannot close file''',
                        2203: '''Cannot open file''',
                        2204: '''Directory not found''',
                        2205: '''File not found''',
                        2206: '''Cannot read current working directory''',
                        2207: '''Cannot change directory''',
                        2208: '''Cannot create directory''',
                        2209: '''Cannot remove directory''',
                        2210: '''File is not a valid script format''',
                        2211: '''File system error''',
                        2212: '''File system command not supported''',
                        2213: '''Too many open files''',
                        2214: '''File access denied''',
                        2215: '''Invalid file handle''',
                        2216: '''Invalid drive''',
                        2217: '''File system busy''',
                        2218: '''Disk full''',
                        2219: '''File corrupt''',
                        2220: '''File already exists''',
                        2221: '''File seek error''',
                        2222: '''End-of-file error''',
                        2223: '''Directory not empty''',
                        2401: '''Invalid specified connection''',
                        2402: '''TSPnet remote error: %s, where %s explains the remote error''',
                        2403: '''TSPnet failure''',
                        2404: '''TSPnet read failure''',
                        2405: '''TSPnet read failure, aborted''',
                        2406: '''TSPnet read failure, timeout''',
                        2407: '''TSPnet write failure''',
                        2408: '''TSPnet write failure, aborted''',
                        2409: '''TSPnet write failure, timeout''',
                        2410: '''TSPnet max connections reached''',
                        2411: '''TSPnet connection failed''',
                        2412: '''TSPnet invalid termination''',
                        2413: '''TSPnet invalid reading buffer table''',
                        2414: '''TSPnet invalid reading buffer index range''',
                        2415: '''TSPnet feature only supported on TSP connections''',
                        2416: '''TSPnet must specify both port and init''',
                        2417: '''TSPnet disconnected by other side''',
                        4900: '''Reading buffer index %s is invalid''',
                        4901: '''The maximum index for this buffer is %d''',
                        4903: '''Reading buffer expired''',
                        4904: '''ICX parameter count mismatch, %s (Line #%d)''',
                        4905: '''ICX parameter invalid value, %s (Line #%d)''',
                        4906: '''ICX invalid function id, %s (Line #%d)''',
                        5001: '''SMU is unresponsive''',
                        5003: '''Saved calibration constants corrupted''',
                        5004: '''Operation conflicts with CALA sense mode''',
                        5005: '''Value too big for range''',
                        5007: '''Operation would exceed safe operating area of the instrument''',
                        5008: '''Operation not permitted while output is on''',
                        5009: '''Unknown sourcing function''',
                        5010: '''No such SMU function''',
                        5011: '''Operation not permitted while cal is locked''',
                        5012: '''Cal data not saved - save or restore before lock''',
                        5013: '''Cannot save cal data - unlock before save''',
                        5014: '''Cannot restore cal data - unlock before restore''',
                        5015: '''Save to cal set disallowed''',
                        5016: '''Cannot change cal date - unlock before operation''',
                        5017: '''Cannot change cal constants - unlock before operation''',
                        5018: '''Cal version inconsistency''',
                        5019: '''Cannot unlock - invalid password''',
                        5021: '''Cannot restore default calset. Using previous calset''',
                        5022: '''Cannot restore previous calset. Using factory calset''',
                        5023: '''Cannot restore factory calset. Using nominal calset''',
                        5024: '''Cannot restore nominal calset. Using firmware defaults''',
                        5025: '''Cannot set filter.count > 1 when measure.count > 1''',
                        5027: '''Unlock cal data with factory password''',
                        5028: '''Cannot perform requested operation while source autorange is enabled''',
                        5029: '''Cannot save without changing cal date and cal due values''',
                        5032: '''Cannot change this setting unless buffer is cleared''',
                        5033: '''Reading buffer not found within device''',
                        5038: '''Index exceeds maximum reading''',
                        5040: '''Cannot use same reading buffer for multiple overlapped measurements''',
                        5041: '''Output Enable not asserted''',
                        5042: '''Invalid while overlapped measure''',
                        5043: '''Cannot perform requested operation while voltage measure autorange is enabled''',
                        5044: '''Cannot perform requested operation while current measure autorange is enabled''',
                        5045: '''Cannot perform requested operation while filter is enabled''',
                        5046: '''SMU too hot''',
                        5047: '''Minimum timestamp resolution is 1μs''',
                        5048: '''Contact check not valid with HIGH-Z output off''',
                        5049: '''Contact check not valid while an active current source''',
                        5050: '''I limit too low for contact check''',
                        5051: '''Model number/SMU hardware mismatch. Disconnect DUT and cycle power''',
                        5052: '''Interlock engaged; system stabilizing''',
                        5053: '''Unstable output detected - Measurements may not be valid''',
                        5054: '''High C voltage limit exceeded''',
                        5055: '''Cannot change adjustment date - change cal constants before operation''',
                        5059: '''trigger.source.action enabled without configuration''',
                        5060: '''trigger.measure.action enabled without configuration''',
                        5061: '''Operation not permitted while OUTPUT is off''',
                        5063: '''Cannot perform requested operation while measure autozero is on''',
                        5064: '''Cannot use reading buffer that collects source values''',
                        5065: '''I range too low for contact check''',
                        5066: '''source.offlimiti too low for contact check''',
                        5069: '''Autorange locked for HighC mode'''}
        # reset
        if reset:
            self.reset()
        else:
            self.get_all()
    
    def _write(self, cmd):
        """
        Sends a visa command <cmd>, waits until "operation complete" and raises eventual errors of the Device.
        
        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.
        
        Returns
        -------
        None
        """
        self._visainstrument.write(cmd)
        while not bool(int(self._visainstrument.query('*OPC?'))):
            time.sleep(1e-6)
        self._raise_error()
        return
    
    def _ask(self, cmd):
        """
        Sends a visa command <cmd>, waits until "operation complete", raises eventual errors of the Device and returns the read answer <ans>.
        
        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.
        
        Returns
        -------
        answer: str
            Answer that is returned at query after the sent <cmd>.
        """
        ans = self._visainstrument.query('print({:s})'.format(cmd)).strip()
        while not bool(int(self._visainstrument.query('*OPC?'))):
            time.sleep(1e-6)
        self._raise_error()
        return ans
    
    def set_digio(self, line, status):
        """
        Sets digital Input/Output line <line> to status <status>. 0 (Off) is 0V and 1 (On) is 5V.
        
        Parameters
        ----------
        line: int
            Number of digital I/O line of interest. Must be in [1,14]
        status: bool
            Output status of selected digital I/O line.
        
        Returns
        -------
        None
        """
        # Corresponding Command: digio.writebit(N, data)
        try:
            logging.debug('{!s}: Set digital I/O of line {:d} to {:d}'.format(__name__, line, status))
            self._write('digio.writebit({:d}, {:d})'.format(line, status))
        except Exception as e:
            logging.error('{!s}: Cannot set digital I/O of line {!s} to {!s}'.format(__name__, line, status))
            raise type(e)('{!s}: Cannot set digital I/O of line {!s} to {!s}\n{!s}'.format(__name__, line, status, e))
        return
    
    def get_digio(self, line):
        """
        Gets digital Input/Output status of line <line>. 0 (Off) is 0V and 1 (On) is 5V.
        
        Parameters
        ----------
        line: int
            Number of digital I/O line of interest. Must be in [1,14]
        
        Returns
        -------
        status: bool
            Output status of selected digital I/O line.
        """
        # Corresponding Command: data = digio.readbit(N)
        try:
            logging.debug('{:s}: Get digital I/O of line {:d}'.format(__name__, line))
            return bool(int(float(self._ask('digio.readbit({:d})'.format(line)))))
        except Exception as e:
            logging.error('{!s}: Cannot get digital I/O of line {!s}'.format(__name__, line))
            raise type(e)('{!s}: Cannot get digital I/O of line {!s}\n{!s}'.format(__name__, line, e))
    
    def set_measurement_mode(self, mode, channel=1):
        """
        Sets measurement mode (wiring system) of channel <channel> to <val>.
        
        Parameters
        ----------
        mode: int
            State of the measurement sense mode. Must be 0 (2-wire), 1 (4-wire) or 3 (calibration).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.sense = senseMode
        try:
            logging.debug('{!s}: Set measurement mode of channel {:s} to {:d}'.format(__name__, chr(64+channel), mode))
            self._write('smu{:s}.sense = {:d}'.format(chr(96+channel), mode))
        except Exception as e:
            logging.error('{!s}: Cannot set measurement mode of channel {!s} to {!s}'.format(__name__, chr(64+channel), mode))
            raise type(e)('{!s}: Cannot set measurement mode of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), mode, e))
        return
    
    def get_measurement_mode(self, channel=1):
        """
        Gets measurement mode (wiring system) <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: int
            State of the measurement sense mode. Meanings are 0 (2-wire), 1 (4-wire) and 3 (calibration).
        """
        # Corresponding Command: senseMode= smuX.sense
        try:
            logging.debug('{!s}: Get measurement mode of channel {:s}'.format(__name__, chr(64+channel)))
            return int(float(self._ask('smu{:s}.sense'.format(chr(96+channel)))))
        except Exception as e:
            logging.error('{!s}: Cannot get measurement mode of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get measurement mode of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_bias_mode(self, mode, channel=1):
        """
        Sets bias mode of channel <channel> to mode <mode>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: int
            Bias mode. Must be 0 (current) or 1 (voltage).
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.func = 0|1|smuX.OUTPUT_DCAMPS|smuX.OUTPUT_DCVOLTS
        try:
            logging.debug('{:s}: Set bias mode of channel {:s} to {:d}'.format(__name__, chr(64+channel), mode))
            self._write('smu{:s}.source.func = {:d}'.format(chr(96+channel), mode))
        except Exception as e:
            logging.error('{!s}: Cannot set bias mode of channel {!s} to {!s}'.format(__name__, chr(64+channel), mode))
            raise type(e)('{!s}: Cannot set bias mode of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), mode, e))
        return
    
    def get_bias_mode(self, channel=1):
        """
        Gets bias mode <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        # Corresponding Command: 0|1|smuX.OUTPUT_DCAMPS|smuX.OUTPUT_DCVOLTS = smuX.source.func
        try:
            logging.debug('{:s}: Get bias mode of channel {:s}'.format(__name__, chr(64+channel)))
            return int(float(self._ask('smu{:s}.source.func'.format(chr(96+channel)))))
        except Exception as e:
            logging.error('{!s}: Cannot get bias mode of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get bias mode of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_sense_mode(self, mode, channel=1):
        """
        Sets sense mode of channel <channel> to mode <mode>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: int
            Sense mode. Must be 0 (current), 1 (voltage), 2 (resistance) or 3 (power).
        
        Returns
        -------
        None
        """
        # Corresponding Command: display.smuX.measure.func = 0|1|2|3|display.MEASURE_DCAMPS|display.MEASURE_DCVOLTS|display.MEASURE_OHMS|display.MEASURE_WATTS
        try:
            logging.debug('{:s}: Set sense mode of channel {:s} to {:d}'.format(__name__, chr(64+channel), mode))
            self._write('display.smu{:s}.measure.func = {:d}'.format(chr(96+channel), mode))
        except Exception as e:
            logging.error('{!s}: Cannot set sense mode of channel {!s} to {!s}'.format(__name__, chr(64+channel), mode))
            raise type(e)('{!s}: Cannot set sense mode of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), mode, e))
        return
    
    def get_sense_mode(self, channel=1):
        """
        Gets sense mode <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: int
            Sense mode. Meanings are 0 (current), 1 (voltage), 2 (resistance) and 3 (power).
        """
        # Corresponding Command: 0|1|2|3|display.MEASURE_DCAMPS|display.MEASURE_DCVOLTS|display.MEASURE_OHMS|display.MEASURE_WATTS = display.smuX.measure.func
        try:
            logging.debug('{:s}: Get sense mode of channel {:s}'.format(__name__, chr(64+channel)))
            return int(float(self._ask('display.smu{:s}.measure.func'.format(chr(96+channel)))))
        except Exception as e:
            logging.error('{!s}: Cannot get sense mode of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense mode of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_bias_range(self, val, channel=1):
        """
        Sets bias range of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias range. Possible values are -1 (auto), for currents 100pA, 1nA, 10nA, 100nA, 1uA, 10uA, 100uA, 1mA, 10mA, 100mA, 1 A, 1.5A and for voltages 200mV, 2V, 20V, 200V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.rangeY = rangeValue
        # Corresponding Command: smuX.source.autorangeY = 0|1|smuX.AUTORANGE_OFF|smuX.AUTORANGE_ON
        try:
            logging.debug('{:s}: Set bias range of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            if val == -1:
                self._write('smu{:s}.source.autorange{:s} = 1'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)]))
            else:
                self._write('smu{:s}.source.autorange{:s} = 0'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)]))
                self._write('smu{:s}.source.range{:s} = {:f}'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)], val))
        except Exception as e:
            logging.error('{!s}: Cannot set bias range of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set bias range of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_bias_range(self, channel=1):
        """
        Gets bias mode <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Bias range.
        """
        # Corresponding Command: rangeValue = smuX.source.rangeY
        try:
            logging.debug('{:s}: Get bias range of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.source.range{:s}'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)])))
        except Exception as e:
            logging.error('{!s}: Cannot get bias range of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get bias range of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_bias_limit(self, val, channel=1):
        """
        Sets bias limit of channel <channel> to <val>.
        Note that the 200V range requires the interlock to be enabled: pull high pin 24 (e.g. by shorting with pin 23) of the digital I/O port.
        
        Parameters
        ----------
        val: float
            Compliance limits. Possible values for currents are 0 < <val> < 1.5A and for voltages are 0 < <val> < 200V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.limitY = limit
        try:
            logging.debug('{:s}: Set bias limit of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.source.limit{:s} = {:f}'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)], val))
        except Exception as e:
            logging.error('{!s}: Cannot set bias limit of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set bias limit of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_bias_limit(self, channel=1):
        """
        Gets bias limit <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Compliance limits.
        """
        # Corresponding Command: limit = smuX.source.limitY
        try:
            logging.debug('{:s}: Get bias limit of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.source.limit{:s}'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)])))
        except Exception as e:
            logging.error('{!s}: Cannot get bias limit of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get bias limit of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_sense_range(self, val, channel=1):
        """
        Sets sense range of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Sense range. Possible values are -1 (auto), for currents 100pA, 1nA, 10nA |100nA, 1uA, 10uA, 100uA, 1mA, 10mA, 100mA, 1 A, 1.5A and for voltages 200mV, 2V, 20V, 200V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.measure.rangeY = rangeValue
        # Corresponding Command: smuX.measure.autorangeY = 0|1|2|smuX.AUTORANGE_OFF|smuX.AUTORANGE_ON|smuX.AUTORANGE_FOLLOW_LIMIT
        try:
            logging.debug('{:s}: Set sense range of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            if val == -1:
                self._write('smu{:s}.measure.autorange{:s} = 1'.format(chr(96+channel), self._smu_command[self.get_sense_mode(channel=channel)]))
            else:
                self._write('smu{:s}.measure.autorange{:s} = 0'.format(chr(96+channel), self._smu_command[self.get_sense_mode(channel=channel)]))
                self._write('smu{:s}.measure.range{:s} = {:f}'.format(chr(96+channel), self._smu_command[self.get_sense_mode(channel=channel)], val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense range of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set sense range of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_sense_range(self, channel=1):
        """
        Gets sense mode <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense range.
        """
        # Corresponding Command: rangeValue = smuX.measure.rangeY
        try:
            logging.debug('{:s}: Get sense range of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.measure.range{:s}'.format(chr(96+channel), self._smu_command[self.get_sense_mode(channel=channel)])))
        except Exception as e:
            logging.error('{!s}: Cannot get sense range of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense range of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_bias_delay(self, val, channel=1):
        """
        Sets bias delay of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias delay with respect to the bias trigger. Possible values are -1 (auto), 0 (off) or positive numbers.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.delay = 0|1|smuX.DELAY_OFF|muX.DELAY_AUTO|sDelay
        try:
            logging.debug('{:s}: Set bias delay of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.source.delay = {:f}'.format(chr(96+channel), val))
        except Exception as e:
            logging.error('{!s}: Cannot set bias delay of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set bias delay of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_bias_delay(self, channel=1):
        """
        Gets bias delay <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Bias delay with respect to the bias trigger.
        """
        # Corresponding Command: 0|1|smuX.DELAY_OFF|muX.DELAY_AUTO|sDelay = smuX.source.delay
        try:
            logging.debug('{:s}: Get bias delay of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.source.delay'.format(chr(96+channel))))
        except Exception as e:
            logging.error('{!s}: Cannot get bias delay of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get bias delay of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_sense_delay(self, val, channel=1, **kwargs):
        """
        Sets sense delay of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Sense delay with respect to the sense trigger. Possible values are -1 (auto), 0 (off) or positive numbers.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        factor: float, optional
            Multiplier to the delays if auto delay is used. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.measure.delay = 0|1|smuX.DELAY_OFF|muX.DELAY_AUTO|mDelay
        # Corresponding Command: smuX.measure.delayfactor = delayFactor
        try:
            logging.debug('{:s}: Set sense delay of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.measure.delay = {:f}'.format(chr(96+channel), val))
            if val == -1:
                self._write('smu{:s}.measure.delayfactor = {:f}'.format(chr(96+channel), kwargs.get('factor', 1)))
        except Exception as e:
            logging.error('{!s}: Cannot set sense delay of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set sense delay of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_sense_delay(self, channel=1):
        """
        Gets sense delay <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense delay with respect to the sense trigger.
        """
        # Corresponding Command: 0|1|smuX.DELAY_OFF|muX.DELAY_AUTO|mDelay= smuX.measure.delay
        try:
            logging.debug('{:s}: Get sense delay of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.measure.delay'.format(chr(96+channel))))
        except Exception as e:
            logging.error('{!s}: Cannot get sense delay of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense delay of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_sense_average(self, val, channel=1, **kwargs):
        """
        Sets sense average of channel <channel> to <val>.
        
        Parameters
        ----------
        val: int
            Number of measured readings that are required to yield one filtered measurement. Must be in [1, 100].
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: str
            Type of filter used for measurements when the measurement filter is enabled. Possible values are 0 (moving average) | 1 (repeat average) | 2 (median). Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.measure.filter.count = filterCount
        # Corresponding Command: smuX.measure.filter.enable = 0|1|smuX.FILTER_OFF|smuX.FILTER_ON
        # Corresponding Command: smuX.measure.filter.type = 0|1|2|smuX.FILTER_MOVING_AVG|smuX.FILTER_REPEAT_AVG|smuX.FILTER_MEDIAN
        try:
            mode = kwargs.get('mode', 1)
            logging.debug('{:s}: Set sense average of channel {:s} to {:d} and mode {:s}'.format(__name__, chr(64+channel), val, self._avg_types[mode]))
            status = not(.5*(1-np.sign(val-1)))  # equals Heaviside(1-<val>) --> turns on for <val> >= 2
            self._write('smu{:s}.measure.filter.enable = {:d}'.format(chr(96+channel), status))
            if status:
                self._write('smu{:s}.measure.filter.count = {:d}'.format(chr(96+channel), val))
                self._write('smu{:s}.measure.filter.type = {:d}'.format(chr(96+channel), mode))
        except Exception as e:
            logging.error('{!s}: Cannot set sense average of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set sense average of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_sense_average(self, channel=1):
        """
        Gets sense average status <status>, value <val> and mode <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        status: bool
            Status if the measurement filter is enabled.
        val: int
            Number of measured readings that are required to yield one filtered measurement.
        mode: str
            Type of filter used for measurements when the measurement filter is enabled. Meanings are 0 (moving average) | 1 (repeat average) | 2 (median).
        """
        # Corresponding Command: filterCount = smuX.measure.filter.count
        # Corresponding Command: 0|1|smuX.FILTER_OFF|smuX.FILTER_ON = smuX.measure.filter.enable
        # Corresponding Command: 0|1|2|smuX.FILTER_MOVING_AVG|smuX.FILTER_REPEAT_AVG|smuX.FILTER_MEDIAN = smuX.measure.filter.type
        try:
            logging.debug('{:s}: Get sense average of channel {:s}'.format(__name__, chr(64+channel)))
            status = bool(int(float(self._ask('smu{:s}.measure.filter.enable'.format(chr(96+channel))))))
            val = int(float(self._ask('smu{:s}.measure.filter.count'.format(chr(96+channel)))))
            mode = int(float(self._ask('smu{:s}.measure.filter.type'.format(chr(96+channel)))))
            return status, val, mode
        except Exception as e:
            logging.error('{!s}: Cannot get sense average of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense average of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_plc(self, val):
        """
        Sets power line cycle (PLC) to <val>.
        
        Parameters
        ----------
        plc: int
            Power line frequency setting used for NPLC calculations. Possible values are -1 (auto) | 50 | 60.
        
        Returns
        -------
        None
        """
        # Corresponding Command: localnode.linefreq = frequency
        # Corresponding Command: localnode.autolinefreq = flag
        try:
            logging.debug('{:s}: Set PLC to {:s}'.format(__name__, str(val)))
            cmd = {-1: 'autolinefreq = true', 50: 'linefreq = 50', 60: 'linefreq = 60'}
            self._write('localnode.{:s}'.format(cmd[int(val)]))
        except Exception as e:
            logging.error('{!s}: Cannot set PLC to {!s}'.format(__name__, val))
            raise type(e)('{!s}: Cannot set PLC to {!s}\n{!s}'.format(__name__, val, e))
        return
    
    def get_plc(self):
        """
        Gets power line cycle (PLC) <val>.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        plc: int
            Power line frequency setting used for NPLC calculations.
        """
        # Corresponding Command: frequency = localnode.linefreq
        # Corresponding Command: flag = localnode.autolinefreq
        try:
            logging.debug('{:s}: Get PLC')
            return float(self._ask('localnode.linefreq'))
        except Exception as e:
            logging.error('{!s}: Cannot get PLC'.format(__name__))
            raise type(e)('{!s}: Cannot get PLC\n{!s}'.format(__name__, e))
    
    def set_sense_nplc(self, val, channel=1):
        """
        Sets sense nplc (number of power line cycle) of channel <channel> with the <val>-fold of one power line cycle.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        val: float
            Integration aperture for measurements. Must be in [0.001, 25].
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.measure.nplc = nplc
        try:
            logging.debug('{:s}: Set sense nplc of channel {:s} to {:f} PLC'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.measure.nplc = {:f}'.format(chr(96+channel), val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense nplc of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set sense nplc of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_sense_nplc(self, channel=1):
        """
        Gets sense nplc (number of power line cycle) <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Integration aperture for measurements.
        """
        # Corresponding Command: nplc = smuX.measure.nplc
        try:
            logging.debug('{:s}: Get sense nplc of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.measure.nplc'.format(chr(96+channel))))
        except Exception as e:
            logging.error('{!s}: Cannot get sense nplc of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense nplc of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_sense_autozero(self, val, channel=1):
        """
        Sets autozero of channel <channel> to <val>. Note that "on" means that the internal zero point is measured for each measurement, wherefore the measurement takes approximately twice as long as when the auto zero function is "off"
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        val: int
            Status of the internal reference measurements (autozero) of the source-measure unit. Possible values are 0 (off), 1 (on) or 2 (once).
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.measure.autozero = azMode
        #                        azMode: 0 (off) | 1 (once) | 2 (at each measurement)
        val = 3*int(bool(val))-val  # swap 1 and 2
        try:
            logging.debug('{:s}: Set sense autozero of channel {:s} to {:d}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.measure.autozero = {:d}'.format(chr(96+channel), val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense autozero of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set sense autozero of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_sense_autozero(self, channel=1):
        """
        Gets sense autozero <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: int
            Status of the internal reference measurements (autozero) of the source-measure unit. Meanings are 0 (off), 1 (on) and 2 (once).
        """
        # Corresponding Command: azMode = smuX.measure.autozero
        #                        azMode: 0 (off) | 1 (once) | 2 (at each measurement)
        try:
            logging.debug('{:s}: Get sense autozero of channel {:s}'.format(__name__, chr(64+channel)))
            val = int(float(self._ask('smu{:s}.measure.autozero'.format(chr(96+channel)))))
            return 3*int(bool(val))-val  # swap 1 and 2
        except Exception as e:
            logging.error('{!s}: Cannot get sense autozero of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense autozero of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_status(self, status, channel=1):
        """
        Sets output status of channel <channel> to <status>.
        
        Parameters
        ----------
        status: int
            Output status. Possible values are 0 (off), 1 (on) or 2 (high Z).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.output = 0|1|2|smuX.OUTPUT_OFF|smuX.OUTPUT_ON|smuX.OUTPUT_HIGH_Z
        try:
            logging.debug('{:s}: Set output status of channel {:s} to {:d}'.format(__name__, chr(64+channel), status))
            self._write('smu{:s}.source.output = {:d}'.format(chr(96+channel), status))
        except Exception as e:
            logging.error('{!s}: Cannot set output status of channel {!s} to {!s}'.format(__name__, chr(64+channel), status))
            raise type(e)('{!s}: Cannot set output status of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), status, e))
        return
    
    def get_status(self, channel=1):
        """
        Gets output status <status> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        status: int
            Output status. Meanings are 0 (off), 1 (on) and 2 (high Z).
        """
        # Corresponding Command: 0|1|2|smuX.OUTPUT_OFF|smuX.OUTPUT_ON|smuX.OUTPUT_HIGH_Z = smuX.source.output
        try:
            logging.debug('{:s}: Get output status of channel {:s}'.format(__name__, chr(64+channel)))
            return int(float(self._ask('smu{:s}.source.output'.format(chr(96+channel)))))
        except Exception as e:
            logging.error('{!s}: Cannot get output status of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get output status of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_bias_value(self, val, channel=1):
        """
        Sets bias value of channel <channel> to value <val>.
        
        Parameters
        ----------
        val: float
            Bias value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.levelY = sourceLevel
        try:
            logging.debug('{:s}: Set bias value of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.source.level{:s} = {:f}'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)], val))
        except Exception as e:
            logging.error('{!s}: Cannot set bias value of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set bias value of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_bias_value(self, channel=1):
        """
        Gets bias value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Bias value.
        """
        # Corresponding Command: sourceLevel = smuX.source.levelY
        try:
            logging.debug('{:s}: Get bias value of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.source.level{:s}'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)])))
        except Exception as e:
            logging.error('{!s}: Cannot get bias value of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get bias value of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
        return
    
    def get_sense_value(self, channel=1, readingbuffer=None):
        """
        Gets sense value of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        readingbuffer: str
            Reading buffer object where all readings will be stored
        
        Returns
        -------
        val: float
            Sense value.
        """
        # Corresponding Command: reading= smuX.measure.Y()
        # Corresponding Command: reading = smuX.measure.Y(readingbuffer)
        try:
            logging.debug('{:s}: Get sense value of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.measure.{:s}({:s})'.format(chr(96+channel), self._smu_command[self.get_sense_mode(channel=channel)], readingbuffer if readingbuffer is not None else '')))
        except Exception as e:
            logging.error('{!s}: Cannot get sense value of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get sense value of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_voltage(self, val, channel=1):
        """
        Sets voltage value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias voltage value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.levelY = sourceLevel
        try:
            logging.debug('{:s}: Set voltage value of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.source.levelv = {:f}'.format(chr(96+channel), val))
        except Exception as e:
            logging.error('{!s}: Cannot set voltage value of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set voltage value of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_voltage(self, channel=1, readingbuffer=None):
        """
        Gets voltage value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        readingbuffer: str
            Reading buffer object where all readings will be stored
        
        Returns
        -------
        val: float
            Sense voltage value.
        """
        # Corresponding Command: reading = smuX.measure.Y()
        # Corresponding Command: reading = smuX.measure.Y(readingbuffer)
        try:
            logging.debug('{:s}: Get voltage value of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.measure.v({:s})'.format(chr(96+channel), readingbuffer if readingbuffer is not None else '')))
        except Exception as e:
            logging.error('{!s}: Cannot get voltage value of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get voltage value of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def set_current(self, val, channel=1):
        """
        Sets current value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias current value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.source.levelY = sourceLevel
        try:
            logging.debug('{:s}: Set current value of channel {:s} to {:f}'.format(__name__, chr(64+channel), val))
            self._write('smu{:s}.source.leveli = {:f}'.format(chr(96+channel), val))
        except Exception as e:
            logging.error('{!s}: Cannot set current value of channel {!s} to {!s}'.format(__name__, chr(64+channel), val))
            raise type(e)('{!s}: Cannot set current value of channel {!s} to {!s}\n{!s}'.format(__name__, chr(64+channel), val, e))
        return
    
    def get_current(self, channel=1, readingbuffer=None):
        """
        Gets current value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        readingbuffer: str
            Reading buffer object where all readings will be stored
        
        Returns
        -------
        val: float
            Sense current value.
        """
        # Corresponding Command: reading = smuX.measure.Y()
        # Corresponding Command: reading = smuX.measure.Y(readingbuffer)
        try:
            logging.debug('{:s}: Get current value of channel {:s}'.format(__name__, chr(64+channel)))
            return float(self._ask('smu{:s}.measure.i({:s})'.format(chr(96+channel), readingbuffer if readingbuffer is not None else '')))
        except Exception as e:
            logging.error('{!s}: Cannot get current value of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get current value of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
        return
    
    def get_IV(self, channel=1, *readingbuffer):
        """
        Gets both current value <I_val> and voltage value <V_val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        readingbuffer: str
            Reading buffer object where current (1st entry) and voltage (2nd entry) readings will be stored.
        
        Returns
        -------
        I_val: float
            Sense current value.
        V_val: float
            Sense voltage value.
        """
        # Corresponding Command: iReading, vReading= smuX.measure.iv()
        # Corresponding Command: iReading, vReading = smuX.measure.iv(ireadingbuffer)
        # Corresponding Command: iReading, vReading = smuX.measure.iv(ireadingbuffer, vreadingbuffer)
        try:
            logging.debug('{:s}: Get current and voltage sense value of channel {:s}'.format(__name__, chr(64+channel)))
            return np.array([float(val) for val in self._ask('smu{:s}.measure.iv{:s}'.format(chr(96+channel), readingbuffer).replace(',)', ')')).split('\t')])
        except Exception as e:
            logging.error('{!s}: Cannot get current and voltage sense value of channel {!s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot get current and voltage sense value of channel {!s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def ramp_bias(self, stop, step, step_time=0.1, channel=2):
        """
        Ramps bias value of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time>.
        
        Parameters
        ----------
        stop: float
            Stop value
        step: float
            Step size.
        step_time: float
            Sleep time between staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1. (default)
        
        Returns
        -------
        None
        """
        start = self.get_bias_value(channel=channel)
        if stop < start:
            step = -step
        for val in np.arange(start, stop, step)+step:
            self.set_bias_value(val, channel=channel)
            print('{:f}{:s}'.format(val, self._IV_units[self.get_bias_mode()]), end='\r')
            time.sleep(step_time)
    
    def ramp_voltage(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps voltage of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time> according to bias_mode.
        
        Parameters
        ----------
        stop: float
            Stop voltage value
        step: float
            Step voltage size.
        step_time: float
            Sleep time between voltage staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1. (default)
        
        Returns
        -------
        None
        """
        if self.get_bias_mode(channel=channel):  # 1 (voltage bias)
            return self.ramp_bias(stop=stop, step=step, step_time=step_time, channel=channel)
        elif not self.get_bias_mode(channel=channel):  # 0 (current bias)
            logging.error(__name__ + ': Cannot set voltage value of channel {!s}: in the current bias'.format(channel))
            return
    
    def ramp_current(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps current of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time> according to bias_mode.
        
        Parameters
        ----------
        stop: float
            Stop current value
        step: float
            Step current size.
        step_time: float
            Sleep time between current staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1. (default)
        
        Returns
        -------
        None
        """
        if not self.get_bias_mode(channel=channel):  # 0 (current bias)
            return self.ramp_bias(stop=stop, step=step, step_time=step_time, channel=channel)
        elif self.get_bias_mode(channel=channel):  # 1 (voltage bias)
            logging.error(__name__ + ': Cannot ramp current value of channel {!s}: in the voltage bias'.format(channel))
    
    def set_sweep_mode(self, mode=1, *channels):
        """
        Sets an variable to decide weather
         * voltage is both applied and measured (VV-mode),
         * current is applied and voltage is measured (IV-mode) or
         * voltage is applied and current is measured (VI-mode)
        and channels <channels> that are used during sweeps.
        
        Parameters
        ----------
        mode: int
            Sweep mode denoting bias and sense modes. Possible values are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode). Default is 1
        channels: int or tuple(int)
            Number of channel of usage. Must be 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).
            
        
        Returns
        -------
        None
        """
        default_channels = {0: (1, 2), 1: (1,), 2: (1,)}
        if mode in [0, 1, 2]:
            if len(channels) == int(not mode)+1:
                logging.debug('{!s}: Set sweep channels to {!s}'.format(__name__, channels))
                self._sweep_channels = channels
            elif len(channels) == 0:
                logging.debug('{!s}: Set sweep channels to {!s}'.format(__name__, default_channels[mode]))
                self._sweep_channels = default_channels[mode]
            else:
                logging.error('{!s}: Cannot set sweep channels to {!s}'.format(__name__, channels))
                raise ValueError('{!s}: Cannot set sweep channels to {!s}'.format(__name__, channels))
            logging.debug('{!s}: Set sweep mode to {:d}'.format(__name__, mode))
            self._sweep_mode = mode
            self.__set_defaults_docstring()
        else:
            logging.error('{!s}: Cannot set sweep mode to {!s}'.format(__name__, mode))
            raise ValueError('{!s}: Cannot set sweep mode to {!s}'.format(__name__, mode))
        return
    
    def get_sweep_mode(self):
        """
        Gets an internal variable <mode> to decide weather
         * voltage is both applied and measured (VV-mode),
         * current is applied and voltage is measured (IV-mode) or
         * voltage is applied and current is measured (VI-mode).
        
        Parameters
        ----------
        None
        
        Returns
        -------
        mode: int
            Sweep mode denoting bias and sense modes. Meanings are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode).
        """
        return self._sweep_mode
    
    def get_sweep_channels(self):
        """
        Gets channels <channels> that are used during sweeps. In the IV and VI sweep mode one channel requires for biasing and sensing, whereas two channels are returned in the VV sweep mode (1st as bias and 2nd as sense channel).
        
        Parameters
        ----------
        None
        
        Returns
        -------
        channels: int or tuple(int)
            Number of channel of usage. Meanings are 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).
        """
        return self._sweep_channels
    
    def get_sweep_bias(self):
        """
        Calls get_bias_mode of channel <channel>. This method is needed for qkit.measure.transport.transport.py in case of no virtual tunnel electronic.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        return self.get_bias_mode(self._sweep_channels[0])
    
    def create_readingbuffer(self, name, size, timestamps=None, channel=1):
        """
        Creates a new reading buffer with name <name> and maximal size <size>.
        
        Parameters
        ----------
        name: str
            Name of the created reading buffer.
        size: int
            Maximum number of readings that can be stored. Maximal value is for default reading buffers is 60 000, when the timestamps and source values options are enabled or over 140 000, when the timestamps and source values options are disabled.
        timestamps: bool
            Status if timestamp values are stored with the readings in the buffer. Default is None.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        """
        # Corresponding Command: bufferVar = smuX.makebuffer(bufferSize)
        # Corresponding Command: bufferVar.collecttimestamps = state
        try:
            logging.debug('{:s}: create reading buffer {:s} for channel {:s} with size {:d}'.format(__name__, name, chr(64+channel), size))
            self._write('{:s} = smu{:s}.makebuffer({:d})'.format(name, chr(96+channel), size))
        except Exception as e:
            logging.error('{:s}: Cannot create reading buffer {:s} for channel {:s} with size {:d}'.format(__name__, name, chr(64+channel), size))
            raise type(e)('{:s}: Cannot create reading buffer {:s} for channel {:s} with size {:d}'.format(__name__, name, chr(64+channel), size, e))
        if timestamps is not None:
            try:
                logging.debug('{:s}: Set timestamps of reading buffer {:s} to {!r}'.format(__name__, name, timestamps))
                self._write('{:s}.collecttimestamps = {!d}'.format(name, timestamps))
            except Exception as e:
                logging.error('{:s}: Cannot set timestamps of reading buffer {:s} to {!r}'.format(__name__, name, timestamps))
                raise type(e)('{:s}: Cannot set timestamps of reading buffer {:s} to {!r}'.format(__name__, name, timestamps, e))
        return
    
    def set_sweep_readingbuffer(self, readingbuffer_bias, readingbuffer_sense):
        """
        Sets reading buffer that is used during sweeps.
        
        Parameters
        ----------
         readingbuffer_bias: str
             Reading buffer object where all bias readings will be stored.
         readingbuffer_sense: str
             Reading buffer object where all sense readings will be stored.
        
        Returns
        -------
        None
        """
        self._readingbuffer_bias = readingbuffer_bias
        self._readingbuffer_sense = readingbuffer_sense
        return
    
    def set_sweep_parameters(self, sweep):
        """
        Sets sweep parameters <sweep> and prepares instrument for the set sweep mode.
        
        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)
        
        Returns
        -------
        None
        """
        # Corresponding Command: format.data = value
        # Corresponding Command: format.asciiprecision = precision
        # Corresponding Command: smuX.trigger.source.linearY(startValue, endValue, points)
        # Corresponding Command: smuX.trigger.source.action = action
        # Corresponding Command: smuX.trigger.measure.action = action
        # Corresponding Command: bufferVar.clear()
        # Corresponding Command: smuX.trigger.measure.iv(ibuffer, vbuffer)
        # Corresponding Command: smuX.trigger.count = triggerCount
        # Corresponding Command: smuX.trigger.arm.count = triggerArmCount
        try:
            self._write('format.data = format.ASCII')
            self._write('format.asciiprecision = 6')
            self._start, self._stop, self._step = np.array(sweep[:3], dtype=float)
            if not self._sweep_mode:  # 0 (VV-mode)
                channel_bias, channel_sense = self._sweep_channels
                readingbuffer_bias = 'smu{:s}.nvbuffer1'.format(chr(96+channel_bias)) if self._readingbuffer_bias is None else self._readingbuffer_bias
                readingbuffer_sense = 'smu{:s}.nvbuffer1'.format(chr(96+channel_sense)) if self._readingbuffer_sense is None else self._readingbuffer_sense
                self._step_signed = np.sign(self._stop-self._start)*np.abs(self._step)
                self.set_bias_value(val=self._start, channel=channel_bias)
                cmd =  '{:s}.clear()'.format(readingbuffer_bias)
                cmd += '{:s}.appendmode = 1'.format(readingbuffer_bias)
                cmd += '{:s}.clear()'.format(readingbuffer_sense)
                cmd += '{:s}.appendmode = 1'.format(readingbuffer_sense)
            elif self._sweep_mode in [1, 2]:  # 1 (IV-mode) | 2 (VI-mode)
                channel, = self._sweep_channels
                readingbuffer_bias = 'smu{:s}.nvbuffer1'.format(chr(96+channel)) if self._readingbuffer_bias is None else self._readingbuffer_bias
                readingbuffer_sense = 'smu{:s}.nvbuffer2'.format(chr(96+channel)) if self._readingbuffer_sense is None else self._readingbuffer_sense
                self._nop = int(np.abs((self._stop-self._start)/self._step)+1)
                cmd =  'smu{:s}.trigger.source.linear{:s}({:f}, {:f}, {:d})'.format(chr(96+channel), self._smu_command[self.get_bias_mode(channel=channel)], self._start, self._stop, self._nop)
                cmd += 'smu{:s}.trigger.source.action = 1'.format(chr(96+channel))
                cmd += 'smu{:s}.trigger.measure.action = 1'.format(chr(96+channel))
                cmd += '{:s}.clear()'.format(readingbuffer_bias)
                cmd += '{:s}.clear()'.format(readingbuffer_sense)
                cmd += 'smu{:s}.trigger.measure.iv({:s}, {:s})'.format(chr(96+channel), *[readingbuffer_bias, readingbuffer_sense][::int(2*(1.5-self._sweep_mode))])
                cmd += 'smu{:s}.trigger.count = {:d}'.format(chr(96+channel), self._nop)
                cmd += 'smu{:s}.trigger.arm.count = 1'.format(chr(96+channel))
            else:
                logging.error('{:s}: Cannot set sweep parameters in sweep mode {!s}'.format(__name__, self._sweep_mode))
                raise AttributeError('{:s}: Cannot set sweep parameters in sweep mode {!s}'.format(__name__, self._sweep_mode))
            self._visainstrument.write(cmd)
            self._prepare_stb('status.OSB')
            time.sleep(1e-3)
        except Exception as e:
            logging.error('{!s}: Cannot set sweep parameters of channel {:s} to {!s}'.format(__name__, self._sweep_channels, sweep))
            raise type(e)('{!s}: Cannot set sweep parameters of channel {:s} to {!s}\n{!s}'.format(__name__, self._sweep_channels, sweep, e))
        return
    
    def get_tracedata(self):
        """
        Starts bias sweep with already set parameters and gets trace data of bias <bias_values> and sense <sense_values> in the set sweep mode.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        # Corresponding Command: bufferVar.clear()
        # Corresponding Command: bufferVar.appendmode = state
        # Corresponding Command: smuX.source.levelY = sourceLevel
        # Corresponding Command: reading = smuX.measure.Y(readingbuffer)
        # Corresponding Command: numberOfReadings = bufferVar.n
        # Corresponding Command: smuX.trigger.initiate()
        # Corresponding Command: waitcomplete()
        try:
            if not self._sweep_mode:  # 0 (VV-mode)
                channel_bias, channel_sense = self._sweep_channels
                readingbuffer_bias = 'smu{:s}.nvbuffer1'.format(chr(96+channel_bias)) if self._readingbuffer_bias is None else self._readingbuffer_bias
                readingbuffer_sense = 'smu{:s}.nvbuffer1'.format(chr(96+channel_sense)) if self._readingbuffer_sense is None else self._readingbuffer_sense
                # sweep channel_bias and measure channel_sense
                cmd = 'for i = {:f}, {:f}, {:f} do'.format(self._start, self._stop+self._step_signed/2., self._step_signed)
                cmd += '\tsmu{:s}.source.levelv = i'.format(chr(96+channel_bias))
                cmd += '\tsmu{:s}.measure.v({:s})'.format(chr(96+channel_bias), readingbuffer_bias)
                cmd += '\tsmu{:s}.measure.v({:s})'.format(chr(96+channel_sense), readingbuffer_sense)
                cmd += 'end\n'
                self._visainstrument.write(cmd)
            elif self._sweep_mode in [1, 2]:  # 1 (IV-mode) | 2 (VI-mode)
                channel, = self._sweep_channels
                readingbuffer_bias = 'smu{:s}.nvbuffer1'.format(chr(96+channel)) if self._readingbuffer_bias is None else self._readingbuffer_bias
                readingbuffer_sense = 'smu{:s}.nvbuffer2'.format(chr(96+channel)) if self._readingbuffer_sense is None else self._readingbuffer_sense
                # sweep and measure channel
                self._visainstrument.write('smu{:s}.trigger.initiate()'.format(chr(96+channel)))
                self._visainstrument.write('waitcomplete()')
            self._wait_for_stb()
            time.sleep(0.1)
            # read data
            self._visainstrument.write('*CLS')
            self._prepare_stb('status.MAV')
            self._visainstrument.write('printbuffer(1, {:s}.n, {:s}, {:s})'.format(readingbuffer_bias, readingbuffer_bias, readingbuffer_sense))
            self._wait_for_stb()
            data = np.fromstring(string=self._visainstrument.read(), dtype=float, count=-1, sep=',')
            return data[0::2], data[1::2]
        except Exception as e:
            logging.error('{!s}: Cannot take sweep data of channel {:s}'.format(__name__, self._sweep_channels))
            raise type(e)('{!s}: Cannot take sweep data of channel {:s}\n{!s}'.format(__name__, self._sweep_channels, e))
    
    def take_IV(self, sweep):
        """
        Takes IV curve with sweep parameters <sweep> in set sweep mode.
        
        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)
        
        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        self.set_sweep_parameters(sweep=sweep)
        return self.get_tracedata()
    
    def _prepare_stb(self, stb):
        """
        Prepares status bit <stb> for high level query.
        
        Parameters
        ----------
        stb: str
            Status bit. Possible values are status.MSB (measurement summary bit), status.SSB (system summary bit), status.EAV (error available), status.QSB (questionable summary bit), status.MAV (meassage available), status.ESB (event summary bit), status.MSS (master summary bit), status.OSB (operation summary bit)
        
        Returns
        -------
        None
        """
        # Corresponding Command: status.reset()
        # Corresponding Command: status.operation.user.condition = operationRegister
        # Corresponding Command: status.operation.user.enable = operationRegister
        # Corresponding Command: status.operation.user.ptr = operationRegister
        # Corresponding Command: status.operation.enable = operationRegister
        # Corresponding Command: status.request_enable = requestSRQEnableRegister
        cmd = '''status.reset()
                 status.operation.user.condition = 0
                 status.operation.user.enable = status.operation.user.BIT0
                 status.operation.user.ptr = status.operation.user.BIT0
                 status.operation.enable = status.operation.USER
                 status.request_enable = {:s}'''.format(stb)
        return self._visainstrument.write(cmd)
    
    def _wait_for_stb(self, wait_time=1e-1):
        """
        Waits until the status bit occurs.
        
        Parameters
        ----------
        wait_time: float
            Sleep time between status bit queries.
        
        Returns
        -------
        None
        """
        # Corresponding Command: status.operation.user.condition = operationRegister
        self._visainstrument.write('status.operation.user.condition = status.operation.user.BIT0')
        self._read_stb()
        time.sleep(wait_time)
        while self._read_stb() == 0:
            time.sleep(wait_time)
        return
    
    def get_OPC(self):
        """
        Gets condition of standard event register entry "Operation complete" <val>.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        val: bool
            Operation complete of standard event register.
        """
        # Corresponding Command: *OPC
        try:
            logging.debug('''{!s}: Get ccr event "Operation complete"'''.format(__name__))
            return bool(int(self._ask('*OPC')))
        except Exception as e:
            logging.error('{!s}: Cannot get ccr condition "Operation complete"'.format(__name__))
            raise type(e)('{!s}: Cannot get ccr condition "Operation complete"\n{!s}'.format(__name__, e))
    
    def _wait_for_OPC(self):
        """
        Waits until the condition of standard event register entry "Operation complete" is fulfilled.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        while not (self.get_OPC()):
            time.sleep(1e-3)
        return
    
    def set_defaults(self, sweep_mode=None):
        """
        Sets default settings that are:
        default parameters
        
        Parameters
        ----------
        mode: int
            Sweep mode denoting bias and sense modes. Possible values are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode). Default is 1
        
        Returns
        -------
        None
        """
        self.reset()
        # beeper off
        self._write('beeper.enable = 0')
        # distiguish different sweep modes
        if sweep_mode is not None:
            self.set_sweep_mode(sweep_mode)
        # set values
        for i, channel in enumerate(self._sweep_channels):
            for func, param in self._defaults[self._sweep_mode][i].items():
                func(param, channel=channel)
        return
    
    def __set_defaults_docstring(self):
        '''
        Sets docstring of "set_defaults" method with actual default settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        '''
        new = 'sweep mode: {:d}:\n'.format(self._sweep_mode) + ''.join(['\t    channel: {:d}\n'.format(channel) + ''.join(['\t\t{:s}: {!s}\n'.format(key_parameter, val_parameter) for key_parameter, val_parameter in self._defaults[self._sweep_mode][i].items()]) for i, channel in enumerate(self._sweep_channels)])
        self.set_defaults.__func__.__doc__ = self.set_defaults.__func__.__doc__.replace(self._default_str, new)
        self._default_str = new
        return
    
    def get_all(self, channel=1):
        """
        Prints all settings of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        logging.debug('{:s}: Get all'.format(__name__))
        print('measurement mode   = {:s}'.format(self._measurement_modes[self.get_measurement_mode(channel=channel)]))
        print('bias mode          = {:d}'.format(self.get_bias_mode(channel=channel)))
        print('sense mode         = {:d}'.format(self.get_sense_mode(channel=channel)))
        print('bias range         = {:1.0e}{:s}'.format(self.get_bias_range(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
        print('sense range        = {:1.0e}{:s}'.format(self.get_sense_range(channel=channel), self._IV_units[self.get_sense_mode(channel=channel)]))
        print('bias delay         = {:1.3e}s'.format(self.get_bias_delay(channel=channel)))
        print('sense delay        = {:1.3e}s'.format(self.get_sense_delay(channel=channel)))
        print('sense average      = {:f}'.format(self.get_sense_average(channel=channel)[1]))
        print('sense average type = {:s}'.format(self._avg_types[self.get_sense_average(channel=channel)[2]]))
        print('plc                = {:f}Hz'.format(self.get_plc()))
        print('sense nplc         = {:f}'.format(self.get_sense_nplc(channel=channel)))
        print('sense autozero     = {:f}'.format(self.get_sense_autozero(channel=channel)))
        print('status             = {!r}'.format(self.get_status(channel=channel)))
        print('bias value         = {:f}{:s}'.format(self.get_bias_value(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
        print('sense value        = {:f}{:s}'.format(self.get_sense_value(channel=channel), self._IV_units[self.get_sense_mode(channel=channel)]))
        for err in self.get_error():
            print('error\t\t   = {:d}\n\t\t     {:s}\n\t\t     {:d}'.format(err[0], err[1], err[2]))
        return
    
    def reset(self, channel=None):
        """
        Resets the instrument or a single channel to default conditions.
        
        Parameters
        ----------
        channel: int
            Number of channel to be set. Must be None, 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: reset()
        # Corresponding Command: smuX.reset()
        if channel is None:
            try:
                logging.debug('{!s}: Reset instrument to factory settings'.format(__name__))
                self._write('reset()')
                self._write('status.reset()')
            except Exception as e:
                logging.error('{!s}: Cannot reset instrument to factory settings'.format(__name__))
                raise type(e)('{!s}: Cannot reset instrument to factory settings\n{!s}'.format(__name__, e))
        else:
            try:
                logging.debug('{!s}: Reset channel {:s} to factory settings'.format(__name__, chr(64+channel)))
                self._write('smu{:s}.reset()'.format(chr(96+channel)))
            except Exception as e:
                logging.error('{!s}: Cannot reset channel {:s} to factory settings'.format(__name__, chr(64+channel)))
                raise type(e)('{!s}: Cannot reset channel {:s} to factory settings\n{!s}'.format(__name__, chr(64+channel), e))
        return
    
    def abort(self, channel=1):
        """
        Aborts the running command of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel to be set. Must be None, 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: smuX.abort()
        try:
            logging.debug('{:s}: Abort running command of channel {:s}'.format(__name__, chr(64+channel)))
            return self._write('smu{:s}.abort()'.format(chr(96+channel)))
        except Exception as e:
            logging.error('{!s}: Cannot abort running command of channel {:s}'.format(__name__, chr(64+channel)))
            raise type(e)('{!s}: Cannot abort running command of channel {:s}\n{!s}'.format(__name__, chr(64+channel), e))
    
    def get_error(self):
        """
        Gets errors of instrument and removes them from the error queue.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        err: array of str
            Entries from the error queue
        """
        # Corresponding Command: errorCount = errorqueue.count
        # Corresponding Command: errorCode, message, severity, errorNode = errorqueue.next() 
        try:
            logging.debug('{!s}: Get errors of instrument'.format(__name__))
            err = [self._visainstrument.query('print(errorqueue.next())').split('\t') for i in range(int(float(self._visainstrument.query('print(errorqueue.count)'))))]
            err = [[int(float(e[0])), str(e[1]), int(float(e[2]))] for e in err]
            if not err:
                err = [[0, 'no error', 0]]
            return err
        except Exception as e:
            logging.error('{!s}: Cannot get errors of instrument'.format(__name__))
            raise type(e)('{!s}: Cannot get errors of instrument\n{!s}'.format(__name__, e))
    
    def _raise_error(self):
        """
        Gets errors of instrument and as the case may be raises a python error.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        errors = self.get_error()
        if len(errors) == 1 and errors[0][0] is 0:  # no error
            return
        else:
            msg = __name__ + ' raises the following errors:'
            for err in errors:
                msg += '\n{:s} ({:s})'.format(self.err_msg[err[0]], err[1])
            raise ValueError(msg)
    
    def clear_error(self):
        """
        Clears error of instrument.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        # Corresponding Command: errorqueue.clear()
        try:
            logging.debug('{:s}: Clear error'.format(__name__))
            self._write('errorqueue.clear()')
        except AttributeError as e:
            logging.error('{!s}: Cannot clear errors'.format(__name__))
            raise type(e)('{!s}: Cannot clear errors\n{!s}'.format(__name__, e))
        return
    
    def close(self):
        """
        Closes the VISA-instrument to disconnect the instrument.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        try:
            logging.debug('{!s}: Close VISA-instrument'.format(__name__))
            self._visainstrument.close()
        except Exception as e:
            logging.error('{!s}: Cannot close VISA-instrument'.format(__name__))
            raise type(e)('{!s}: Cannot close VISA-instrument\n{!s}'.format(__name__, e))
        return
    
    def get_parameters(self):
        """
        Gets a parameter list <parlist> of measurement specific setting parameters.
        Needed for .set-file in 'write_additional_files', if qt parameters are not used.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        parlist: dict
            Parameter names as keys, corresponding channels of interest as values.
        """
        parlist = {'measurement_mode': {'channels': [1, 2]},
                   'bias_mode': {'channels': [1, 2]},
                   'sense_mode': {'channels': [1, 2]},
                   'bias_range': {'channels': [1, 2]},
                   'sense_range': {'channels': [1, 2]},
                   'bias_delay': {'channels': [1, 2]},
                   'sense_delay': {'channels': [1, 2]},
                   'sense_average': {'channels': [1, 2]},
                   'bias_value': {'channels': [1, 2]},
                   'plc': {'channels': [None]},
                   'sense_nplc': {'channels': [1, 2]},
                   'sense_autozero': {'channels': [1, 2]},
                   'status': {'channels': [1, 2]}}
        return parlist

    def get(self, param, **kwargs):
        """
        Gets the current parameter <param> by evaluation 'get_'+<param> and corresponding channel if needed
        In combination with <self.get_parameters> above.

        Parameters
        ----------
        param: str
            Parameter of interest.
        channels: array_likes
            Number of channels of interest. Must be in {1, 2} for channel specific parameters or None for channel independent (global) parameters.

        Returns
        -------
        parlist: dict
            Parameter names as keys, values of corresponding channels as values.
        """
        channels = kwargs.get('channels').get('channels')
        if channels == [None]:
            return getattr(self, 'get_{!s}'.format(param))()
        else:
            return tuple([getattr(self, 'get_{!s}'.format(param))(channel=channel) for channel in channels])



# -*- coding: utf-8 -*-
# qData.py analysis class for qkit measurements data
# Micha Wildermuth, micha.wildermuth@kit.edu 2023

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import numpy as np
from collections import defaultdict
import itertools

import qkit
from qkit.storage.store import Data
import qkit.measure.measurement_class as mc
from qkit.gui.plot import plot as qviewkit
from qkit.storage import store as hdf
from qkit.storage.hdf_constants import ds_types
import json
from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder


class dict2obj(object):
    def __init__(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
                setattr(self, a.replace(' ', '_'), [dict2obj(x) if isinstance(x, dict) else x for x in b])
            else:
                setattr(self, a.replace(' ', '_'), dict2obj(b) if isinstance(b, dict) else b)


class qData(object):
    """
    This is an analysis class for qkit measurements.
    """

    def __init__(self):
        """
        Initializes an analysis class for qkit .

        Parameters
        ----------
        None

        Returns
        -------
        None

        Examples
        --------
        >>> import numpy as np
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        Initialized the file info database (qkit.fid) in 0.000 seconds.

        >>> from qkit.analysis.qdata import qData
        >>> qd = qData()
        """
        # qkit.fid.update_file_db()  # update file database
        self.uuid = self.path = self.df = None
        self.analysis = self.views = None
        self.settings = None
        self.measurement = mc.Measurement()  # qkit-sample object
        self.m_type = self.scan_dim = None
        self.x_ds = self.x_coordname = self.x_unit = self.x_vec = None
        self.y_ds = self.y_coordname = self.y_unit = self.y_vec = None
        self.si_prefix = {'y': 1e-24,  # yocto
                          'z': 1e-21,  # zepto
                          'a': 1e-18,  # atto
                          'f': 1e-15,  # femto
                          'p': 1e-12,  # pico
                          'n': 1e-9,  # nano
                          'u': 1e-6,  # micro
                          'm': 1e-3,  # milli
                          'c': 1e-2,  # centi
                          'd': 1e-1,  # deci
                          '': 1e0,
                          'k': 1e3,  # kilo
                          'M': 1e6,  # mega
                          'G': 1e9,  # giga
                          'T': 1e12,  # tera
                          'P': 1e15,  # peta
                          'E': 1e18,  # exa
                          'Z': 1e21,  # zetta
                          'Y': 1e24,  # yotta
                          }

    def load(self, uuid):
        """
        Loads qkit measurement data with given uuid <uuid>.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and loaded

        Returns
        -------
        None

        Examples
        --------
        >>> qd.load(uuid='XXXXXX')
        """
        if uuid != self.uuid and self.uuid is not None:
            self.__init__()
        self.uuid = uuid
        self.path = qkit.fid.get(self.uuid)
        self.df = Data(self.path)

        ''' load all entries '''
        for key, val in self.df.__dict__.items():  # all entries in df, except for hf, analysis, data, views, and default_ds
            if key[0] == '_':
                setattr(self, key[1:], val)
        for key, val in self.df.data.__dict__.items():  # all entries in data
            if key == 'settings':
                try:
                    self.settings = dict2obj(json.loads(self.df.data.settings[0], cls=QkitJSONDecoder))
                except AttributeError:
                    self.settings = None
                except:
                    self.settings = self.df.data.settings[:]
            elif key == 'measurement':
                try:
                    self.measurement.load(qkit.fid.measure_db[self.uuid])
                except:
                    self.measurement = dict2obj(json.loads(self.df.data.measurement[0], cls=QkitJSONDecoder))
            else:
                setattr(self, key, val[:])
        self.analysis = dict2obj({key: val[:] for key, val in self.df.analysis.__dict__.items()})  # all entries in analysis
        self.views = dict2obj({key: val for key, val in self.df['entry/views'].items()})  # all entries in views
        for name, view in self.views.__dict__.items():
            for a, b in self.views.__dict__[name].attrs.items():
                if not hasattr(view, a):  # avoid conflicts with same keys
                    setattr(self.views.__dict__[name], a, b)  # set view attributes (e.g. xy_0) as attributes of qd.views.<view>
        self.m_type = self.measurement.measurement_type  # measurement type

    def _get_xy_parameter(self, measurand):
        """
        Identifies eventual x- and y-parameters.

        Parameters
        ----------
        measurand: HDF5 dataset
            Dataset, whose attributes 'x_ds_url' and 'y_ds_url' are used to identify eventual x- and y-parameters.

        Returns
        -------
        None
        """
        if self.scan_dim >= 2:  # 2D or 3D scan
            # x parameter
            self.x_ds = self.df[measurand.attrs['x_ds_url']]
            self.x_coordname = self.x_ds.attrs['name']
            self.x_unit = self.x_ds.attrs['unit']
            self.x_vec = self.x_ds[:]
            if self.scan_dim == 3:  # 3D scan
                # y parameter
                self.y_ds = self.df[measurand.attrs['y_ds_url']]
                self.y_coordname = self.y_ds.attrs['name']
                self.y_unit = self.y_ds.attrs['unit']
                self.y_vec = self.y_ds[:]
            else:
                self.y_ds, self.y_coordname, self.y_unit, self.y_vec = None, None, None, None
        else:
            self.x_ds, self.x_coordname, self.x_unit, self.x_vec = None, None, None, None
        return

    def merge(self, uuids, order=None):
        """
        Merges qkit measurement data of several individual files with given uuids <uuids>.
        * 1D: all sweep data are stacked and views are merged.
        * 2D: values of x-parameter and its corresponding sweep data are merged in the order <order>.
        * 3D: values of x- and y-parameters and its corresponding sweep data are merged in the order <order>.

        Parameters
        ----------
        uuids: list of str
            Qkit identification names, that are looked for, loaded and merged.
        order: list of int (optional)
            Order by which data are merged. It is used to slice the data np.arrays via [::<order>], where 1 means same
            and -1 opposite direction. Default is 1 (same direction).

        Returns
        -------
        uuid: str
            Qkit identification names of created file

        Examples
        --------
        >>> qd.merge(uuid=['XXXXXX', 'YYYYYY'], order=[-1, 1])
        """

        def get_key(_key, _dict):
            """
            Check if key already exists and increases counter by one if so.
            """
            if _key in _dict.keys():
                _parts = _key.split('_')
                _key = '_'.join([str(int(_part) + 1) if _part.isdigit() else _part for _part in _parts])
                return get_key(_key, _dict)
            else:
                return _key

        def remove_duplicates(_list):
            _res = []
            for _elem in _list:
                if _elem not in _res:
                    _res.append(_elem)
            if len(_res) == 1:
                return _res[0]
            else:
                return _res

        def merge_dict(_listdict):
            dict_merged = defaultdict(list)
            for _dict in _listdict:
                for _key, _val in _dict.items():
                    dict_merged[_key].append(_val)
            return dict_merged

        # TODO: write additional files and plots
        order = np.ones_like(uuids, dtype=int) if order is None else np.array(order, dtype=int)
        ''' load data from input files '''
        data = []
        attrs = []
        run_user = []
        scan_dim = []
        for i, uuid in enumerate(uuids):
            data.append({})
            attrs.append({})
            self.load(uuid)
            run_user.append([self.measurement.run_id, self.measurement.user])
            scan_dim.append(self.scan_dim)
            for group in self.df.hf.entry:
                for dataset in self.df.hf.entry[group]:
                    path = '/'.join((group, dataset))
                    data[i][path] = self.df.hf.entry[path][:][::order[i]]
                    attrs[i][path] = self.df.hf.entry[path].attrs.items()
        ''' merge data '''
        if set(scan_dim) == {1}:
            keys, data_merged, attrs_merged = {}, {}, {}
            for j, d in enumerate(data):
                for key, val in d.items():
                    ds_type = dict(attrs[j][key])['ds_type']
                    if ds_type < 5:  # coordinate, vector, matrix or box
                        k = get_key(_key=key, _dict=keys)
                        keys[k] = (key, j)
                        data_merged[k] = [data[j][key]] * 2
                        attrs_merged[k] = (attrs[j][key],)
                    elif ds_type == ds_types['txt']:
                        keys[key] = (key,)
                        data_merged[key] = [d[key] for d in data]
                        attrs_merged[key] = [a[key] for a in attrs]
                    elif ds_type == ds_types['view']:
                        keys[key] = (key,)
                        data_merged[key] = (data[j][key],)
                        if j == 0:
                            attrs_merged[key] = [dict(attrs[j][key])] * 2
                        else:
                            for k, v in dict(attrs[j][key]).items():
                                # increase X of 'xy_X' and its values to the next possible
                                if 'xy_' in k and '_filter' not in k:
                                    new_key = get_key(_key=k, _dict=attrs_merged[key][0])
                                    attrs_merged[key][0][new_key] = ':'.join(
                                        ['_'.join([new_key.strip('xy_') if s.isdigit() else s for s in ulr.split('_')])
                                         for ulr in str(v, encoding='utf-8').split(':')]).encode()
                                # increase X of 'xy_X_filter' to the next possible
                                elif 'xy_' in k and '_filter' in k:
                                    attrs_merged[key][0][get_key(_key=k, _dict=attrs_merged[key][0])] = v
        else:
            keys = {key: (key,) for key in data[0].keys()}
            data_merged = merge_dict(_listdict=data)
            attrs_merged = merge_dict(_listdict=attrs)
        ''' write data to new file '''
        # create new file
        qkit.cfg['run_id'], qkit.cfg['user'] = np.squeeze(np.vstack({tuple(row) for row in run_user}))
        _data_file = hdf.Data(name='+'.join(uuids), mode='a')
        # create and write measurement, settings and coordinates
        ds = {}
        for key, val in data_merged.items():
            name = key.split('/')[1]
            attr = dict(attrs_merged[key][0])
            ds_type = attr['ds_type']
            if ds_type == ds_types['txt']:  # measurement, settings
                txt_merged = merge_dict(_listdict=[json.loads(d[0], cls=QkitJSONDecoder) for d in val])
                ds[key] = _data_file.add_textlist(name)
                if 'settings' in key:
                    ds[key].append({kk: {k: remove_duplicates(_list=map(lambda x: x[k], dic)) for k in dic[0]}
                                    for kk, dic in dict(txt_merged).items()})
                elif 'measurement' in key:
                    if set(scan_dim) == {1}:
                        # merge sweeps in sample object
                        for k, v in txt_merged.items():
                            if np.iterable(v) and type(v[0]) is dict:
                                txt_merged[k] = [{k: np.vstack(v) if 'sweeps' in k else remove_duplicates(_list=v) for
                                                  k, v in merge_dict(_listdict=v).items()}] * 2
                    ds[key].append({k: remove_duplicates(_list=v) for k, v in txt_merged.items()})
            elif ds_type == ds_types['coordinate']:
                ds[key] = _data_file.add_coordinate(name=name,
                                                    unit=str(dict(attrs_merged[keys[key][0]][0])['unit'],
                                                             encoding='utf-8'))
                if np.array([np.array_equal(p[0], p[1])
                             for p in
                             itertools.combinations(np.array(list(itertools.zip_longest(*val, fillvalue=np.nan))).T,
                                                    2)]).all():
                    ds[key].add(val[0])
                else:
                    ds[key].add(np.concatenate(val))
        # create datasets
        for key, val in data_merged.items():
            name = key.split('/')[1]
            attr = dict(attrs_merged[key][0])
            ds_type = attr['ds_type']
            folder = key.split('0/')[0]
            if ds_type == ds_types['vector']:
                ds[key] = _data_file.add_value_vector(name=name,
                                                      x=ds[str(attr['x_ds_url'], encoding='utf-8').strip('/entry')],
                                                      unit=str(attr['unit'], encoding='utf-8'),
                                                      folder=folder,
                                                      save_timestamp=False)
            elif ds_type == ds_types['matrix']:
                ds[key] = _data_file.add_value_matrix(name=name,
                                                      x=ds[str(attr['x_ds_url'], encoding='utf-8').strip('/entry')],
                                                      y=ds[str(attr['y_ds_url'], encoding='utf-8').strip('/entry')],
                                                      unit=str(attr['unit'], encoding='utf-8'),
                                                      folder=folder,
                                                      save_timestamp=False)
            elif ds_type == ds_types['box']:
                ds[key] = _data_file.add_value_box(name=name,
                                                   x=ds[str(attr['x_ds_url'], encoding='utf-8').strip('/entry')],
                                                   y=ds[str(attr['y_ds_url'], encoding='utf-8').strip('/entry')],
                                                   z=ds[str(attr['z_ds_url'], encoding='utf-8').strip('/entry')],
                                                   unit=str(attr['unit'], encoding='utf-8'),
                                                   folder=folder,
                                                   save_timestamp=False)
        # write values to datasets and create and write views
        for key, val in data_merged.items():
            name = key.split('/')[1]
            attr = dict(attrs_merged[key][0])
            ds_type = attr['ds_type']
            if ds_type == ds_types['vector']:
                ds[key].append(val[0])
            elif ds_type == ds_types['matrix']:
                [ds[key].append(x) for x in np.concatenate(val)]
            elif ds_type == ds_types['box']:
                for xs in np.concatenate(val):
                    [ds[key].append(x) for x in xs]
                    ds[key].next_matrix()
            elif ds_type == ds_types['view']:
                if np.array([np.array_equal(p[0], p[1]) for p in
                             itertools.combinations(list(map(list, attrs_merged[key])), 2)]).all():
                    for j, xy in enumerate(
                            filter(lambda k: 'xy_' in k and '_filter' not in k, dict(attrs_merged[key][0]).keys())):
                        x, y = [i.strip('/entry') for i in
                                str(dict(attrs_merged[key][0])[xy], encoding='utf-8').split(':')]
                        if j == 0:
                            ds[key] = _data_file.add_view(name=name, x=ds[x], y=ds[y],
                                                          view_params=json.loads(
                                                              dict(attrs_merged[key][0])['view_params']))
                        else:
                            ds[key].add(x=ds[x], y=ds[y])
        # close new file
        _data_file.close_file()
        return _data_file.__dict__['_uuid']

    def open_qviewkit(self, uuid=None, ds=None):
        """
        Opens qkit measurement data with given uuid <uuid> in qviewkit.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and opened in qviewkit
        ds: str | list(str)
            Datasets that are opened instantaneously.

        Returns
        -------
        None

        See also
        --------
        qkit.gui.plot.plot
            qviewkit gui to display qkit .h5-files.
        """
        if uuid is None:
            path = self.path
        else:
            path = qkit.fid.get(uuid)
        if ds is None or np.iterable(ds) and all(isinstance(x, str) for x in ds):
            pass
        elif type(ds) is str:
            ds = [ds]
        else:
            raise ValueError('Argument <ds> needs to be set properly.')
        qviewkit.plot(path, datasets=ds, live=False)  # opens ds by default






#!/usr/bin/env python2
# server.py by JB@KIT 01/2018

'''
Server script for DC DAC LTC2666 to be run on Raspberry Pi mounted in the DC DAC rack. 
Raspberry Pi 3 model pin configuration: MOSI: 19, MISO: 21, SCLK: 23, Chip select CE0: 24 (in use with dac0), CE1: 26 (dac1). 
DAC model: LTC2666-16

Internal reference: REFCOMP floating, RD bit = 0.
'''

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import numpy as np
import time,sys
import logging

import zerorpc

from DAC_LTC2666_communication import DAC_LTC2666
dacs = []
dac0 = DAC_LTC2666(cs = 0, word32 = True)   #load DAC on CS0 port
dacs.append(dac0)
dac1 = DAC_LTC2666(cs = 1, word32 = True)   #load DAC on CS1 port
dacs.append(dac1)
    

class DAC_Server(object):
    '''
    Main server class.
    '''
    def init_dac(self):
        '''
        Initialize dac boards.
        '''
        for dac in dacs:
            dac.init()
    def verify(self):
        '''
        Verify dac boards by checking acknowledge bytes.
        '''
        for i,dac in enumerate(dacs):
            if not dac.verify_connection():
                return 'No acknowledge received from DAC #{:d}'.format(i)
        return True
    def set_voltage(self, channel, voltage):
        '''
        Main method to set a dac voltage value via the server.
        - channel: integer within 0..15
        - voltgae: voltage to be set: -5V..5V
        '''
        return dacs[int(channel)/8].set_voltage((channel%8), voltage)
       

if __name__ == "__main__":   #if executed as main (and not imported)
    '''
    Start routine for dac server. HOST and PORT can be passed as arguments, otherwise defaults are used.
    '''
    HOST = 'localhost'
    PORT = 9931
    print sys.argv
    try:
        HOST = sys.argv[1]
        PORT = int(sys.argv[2])
    except IndexError:
        print 'Using defaults: {:s}, {:d}'.format(HOST, PORT)
    #time.sleep(0.1) 
    
    s = zerorpc.Server(DAC_Server())
    s.bind("tcp://{:s}:{:d}".format(HOST, PORT))
    print 'Server started...\n'
    s.run()


# Toolset for calibrating and using IQ-Mixers as Single-Sideband-Mixers (SSB)
# Started by Andre Schneider 01/2015 <andre.schneider@student.kit.edu>

import qkit
from qkit.core.instrument_base import Instrument
import time
import types
import logging
import os
import numpy as np
if qkit.module_available("matplotlib"):
    import matplotlib.pyplot as plt
import sys, gc
from copy import copy


class IQ_Mixer(Instrument):

    def __init__(self, name, sample, mixer_name):
        Instrument.__init__(self, name, tags=['virtual'])
        self._sample = sample
        self._FSUP_connected = False
        self._fsup = None
        self._swb = None
        # self._iq_frequency = None
        self._output_power = None
        self._sideband_frequency = None
        self._mw_power = None
        self.maxage = 100 * 24 * 3600  # Age of calibration data in seconds
        self.interpol_freqspan = 1e9  # Frequency span where interpolated data is used and recalibrated
        self.trust_region = 500e6  # Frequency span where interpolated data is used without recalibration. Do not set this to zero, because calibration will round frequency
        self._iq_frequency = self._sample.iq_frequency
        self.do_set_mixer_name(mixer_name)

        self.cache_valid = False
        self.cache_frequency = self._sample.f01
        self.cache_power = self._sample.mw_power
        self.cache_iq = self._sample.iq_frequency

        # Parameters
        self.add_parameter('sideband_frequency', type=float,
                           flags=Instrument.FLAG_GET, units='Hz',
                           minval=1, maxval=20e9)

        self.add_parameter('output_power', type=float,
                           flags=Instrument.FLAG_GET, units='dBm')

        self.add_parameter('iq_frequency', type=float,
                           flags=Instrument.FLAG_GET, units='Hz',
                           minval=0,
                           maxval=1e9)  # This is only a get variable, because changing requires an update of all sequences in AWG

        self.add_parameter('FSUP_connected', type=bool,
                           flags=Instrument.FLAG_GET)

        self.add_parameter('mixer_name', type=str,
                           flags=Instrument.FLAG_GETSET)

        self.add_function('connect_FSUP')
        self.add_function('disconnect_FSUP')
        self.add_function('get_all')
        self.add_function('set_sample')

    IQ0307 = 'IQ0307'  # simply define the currently known mixers so you can see them when pressing tab
    IQ0318 = 'IQ0318'
    IQ4509 = 'IQ4509'

    _ch1_filename = 'ch1_opt'
    _ch2_filename = 'ch2_opt'
    _awg_filepath = 'C:\\waveforms\\'
    (x, y) = (0, 0)  # initial values for DC offset in Volts

    def get_all(self):
        self.get_FSUP_connected()
        self.get_iq_frequency()
        self.get_output_power()
        self.get_sideband_frequency()

    def ch1(self, t):
        return np.cos(t)

    def ch2(self, t):
        return np.sin(t)

    def set_sample(self, sample):
        self._sample = sample
        self.cache_valid = False

    def do_set_mixer_name(self, mixer_name):
        self.mixer_name = mixer_name
        self.cache_valid = False

    def do_get_mixer_name(self):
        return self.mixer_name

    def do_get_FSUP_connected(self):
        return self._FSUP_connected

    def do_set_sideband_frequency(self, frequency):
        self._sideband_frequency = frequency

    def do_get_sideband_frequency(self):
        return self._sideband_frequency

    def do_set_mw_power(self, mw_power):
        self._mw_power = mw_power

    def do_get_output_power(self):
        return self._output_power

    def do_get_iq_frequency(self):
        '''This returns the iq_frequency which was used for the last call of convert()'''
        return self._iq_frequency

    def do_set_iq_frequency(self, iq_frequency):
        self._iq_frequency = iq_frequency

    def connect_FSUP(self, fsup):
        self._fsup = fsup
        self._FSUP_connected = True

    def disconnect_FSUP(self):
        self._FSUP_connected = False

    def connect_switchbox(self, swb):
        '''
        as a default, Ch1 is cryo, Ch2 is fsup.
        '''
        self._swb = swb

    def disconnect_switchbox(self):
        self._swb = None

    def dcoptimize(self, DC):
        # you need to have a zero-waveform loaded in the awg and turn outputs on!
        if (np.abs(DC[0]) > .5 or np.abs(DC[1]) > .5):
            qkit.flow.sleep()
            return 50
        self._sample.awg.set({'ch1_offset': DC[0], 'ch2_offset': DC[1]})
        # qkit.flow.sleep(.05)
        self._fsup.sweep()
        return np.around(self._fsup.get_marker_level(1), 2)  # assuming that marker1 is at leakage frequency

    def xoptimize(self, x):
        # you need to have a zero-waveform loaded in the awg and turn outputs on!
        if (np.abs(x) > .5):
            qkit.flow.sleep()
            return 50
        self._sample.awg.set_ch1_offset(x)
        self._fsup.sweep()
        return np.around(self._fsup.get_marker_level(1), 2)  # assuming that marker1 is at leakage frequency

    def yoptimize(self, y):
        # you need to have a zero-waveform loaded in the awg and turn outputs on!
        if (np.abs(y) > .5):
            qkit.flow.sleep()
            return 50
        self._sample.awg.set_ch2_offset(y)
        self._fsup.sweep()
        return np.around(self._fsup.get_marker_level(1), 2)  # assuming that marker1 is at leakage frequency

    def relampoptimize(self, amp):
        self._sample.awg.set_ch1_amplitude(amp)
        self._fsup.sweep()
        return np.around(self._fsup.get_marker_level(4), 2)  # assuming that marker4 is at unwanted sideband

    def relampoptimize2(self, amp):
        self._sample.awg.set_ch2_amplitude(amp)
        self._fsup.sweep()
        return np.around(self._fsup.get_marker_level(4), 2)  # assuming that marker4 is at unwanted sideband

    def focus(self, frequency, marker):
        self._fsup.enable_marker(1, 'OFF')
        self._fsup.enable_marker(2, 'OFF')
        self._fsup.enable_marker(3, 'OFF')
        self._fsup.enable_marker(4, 'OFF')
        self._fsup.set_continuous_sweep_mode('OFF')
        self._fsup.set({'centerfreq': frequency,
                        'freqspan': 2e2,
                        'resolutionBW': 200,
                        'videoBW': 200})
        self._fsup.set_marker(marker, frequency)
        sweeptime = self._fsup.get_sweeptime()
        self._fsup.sweep()
        return sweeptime

    def load_zeros(self):
        ch1wfm = np.zeros(10)
        ch2wfm = np.zeros(10)
        marker = np.zeros_like(ch1wfm)
        self._sample.awg.wfm_send(ch1wfm, marker, marker, self._awg_filepath + self._ch1_filename, self._sample.clock)
        self._sample.awg.wfm_import(self._ch1_filename, self._awg_filepath + self._ch1_filename, 'WFM')
        self._sample.awg.wfm_send(ch2wfm, marker, marker, self._awg_filepath + self._ch2_filename, self._sample.clock)
        self._sample.awg.wfm_import(self._ch2_filename, self._awg_filepath + self._ch2_filename, 'WFM')
        self._sample.awg.set({'ch1_waveform': self._ch1_filename,
                              'ch1_offset': 0,
                              'ch1_amplitude': 1,
                              'ch1_waveform': self._ch2_filename,
                              'ch1_offset': 0,
                              'ch1_amplitude': 1})
        self._sample.awg.run()
        self._sample.awg.set({'ch1_output': 1, 'ch2_output': 1})

    def load_wfm_init(self):
        t = np.linspace(0, 2 * np.pi, self._sample.clock / self._iq_frequency, endpoint=False)
        ch1wfm = self.ch1(t)
        ch2wfm = self.ch2(t)
        marker = np.zeros_like(ch1wfm)
        self._sample.awg.wfm_send(ch2wfm, marker, marker, self._awg_filepath + self._ch2_filename, self._sample.clock)
        self._sample.awg.wfm_import(self._ch2_filename, self._awg_filepath + self._ch2_filename, 'WFM')
        self._sample.awg.set_ch2_waveform(self._ch2_filename)
        self._sample.awg.wfm_send(ch1wfm, marker, marker, self._awg_filepath + self._ch1_filename, self._sample.clock)
        self._sample.awg.wfm_import(self._ch1_filename, self._awg_filepath + self._ch1_filename, 'WFM')
        self._sample.awg.set_ch1_waveform(self._ch1_filename)
        self._sample.awg.run()
        self._sample.awg.set({'ch1_output': 1, 'ch2_output': 1})

    def optimize_dc(self, frequency, x, y, maxiter=10, verbose=False):
        self.focus(frequency, 1)
        for i in range(0, maxiter):
            xold = x
            yold = y
            x = opt.minimize_scalar(lambda x: self.dcoptimize([x, y]), method='golden', tol=0.1).x
            if (verbose): print(" %.3f  (%.3f) %.2fdBm" % (x, y, self.dcoptimize([x, y])))
            sys.stdout.flush()
            y = opt.minimize_scalar(lambda y: self.dcoptimize([x, y]), method='golden', tol=0.1).x
            if (verbose): print("(%.3f)  %.3f  %.2fdBm" % (x, y, self.dcoptimize([x, y])))
            sys.stdout.flush()
            if (np.around(xold - x, 3) == 0 and np.around(yold - y, 3) == 0):
                if (verbose): print("Finished after %i iterations with delta x: %.5f delta y: %.5f" % (
                    i + 1, xold - x, yold - y))
                break
        return [x, y]

    def optimize_phase(self):
        self.focus(self._f_rounded - 2 * self._iq_frequency, 4)
        return opt.minimize_scalar(self.phaseoptimize, method="bounded", bounds=(0, 2 * np.pi),
                                   options={"xatol": .01}).x

    def optimize_relamp(self):
        self.focus(self._f_rounded - 2 * self._iq_frequency, 4)
        return opt.minimize_scalar(self.relampoptimize, method="bounded", bounds=(0, 2), options={"xatol": .001}).x

    def findmin(self, function, start, stop, stepsize, plot=False, averages=1):
        a = np.arange(start, stop + stepsize / 2, stepsize)
        if (len(a) < 2):
            # print "*************TRYING TO MINIMIZE %s BETWEEN %f and %f => NOT POSSIBLE!*********"%(function,start,stop)
            return start
        b = np.zeros(len(a))
        for u, v in enumerate(a):
            for i in range(averages):
                b[u] += function(v)
        b = b / averages
        if plot and qkit.module_available("matplotlib"):
            plt.plot(a, b, "k*-")
            plt.show()
        return a[b.argmin()]

    def minimize(self, function, start, stop, init_stepsize, min_stepsize, initial_averages=1, final_averages=5,
                 verbose=False, hardbounds=False, confirmonly=False, bounds=(-np.inf, np.inf)):
        '''
            returns (value,function value) where function is minimal
            The bounds are not really hard: if this function leaves your bounds, you should increase initial stepsize and initial averages
        '''
        if hardbounds:
            print("HARDBOUNDS is no longer supported, use bounds=(lowe,upper) instead!")
        if confirmonly:
            stepsize = min_stepsize
            x = (start + stop) / 2
            initial_averages = final_averages
        else:
            stepsize = init_stepsize
            x = self.findmin(function, start, stop, stepsize, verbose, initial_averages)
            if (verbose): print("stepsize was %f, minimum detected at %f" % (init_stepsize, x))
            if (x == start):
                x = start - 3 * stepsize
            elif (x == stop):
                x = stop + 3 * stepsize
            else:
                stepsize = max(stepsize / 10, min_stepsize)
        while True:
            xold = x
            x = self.findmin(function, max(bounds[0], x - 3 * stepsize), min(bounds[1], x + 3 * stepsize), stepsize,
                             verbose, initial_averages)
            if (verbose): print("stepsize was %f, minimum detected at %f" % (stepsize, x))
            if (x == xold - 3 * stepsize):
                x -= 2 * stepsize
            elif (x == xold + 3 * stepsize):
                x += 2 * stepsize
            else:
                if confirmonly:
                    return (x, function(x))
                if (stepsize == min_stepsize):
                    break
                stepsize = max(stepsize / 5, min_stepsize)
        if (verbose): print("Finishing with stepsize %f" % (stepsize))
        x = self.findmin(function, max(bounds[0], x - 2 * stepsize), min(bounds[1], x + 2 * stepsize), stepsize,
                         verbose, final_averages)
        return (x, function(x))

    def load_zeros(self):
        ch1wfm = np.zeros(10)
        ch2wfm = np.zeros(10)
        marker = np.zeros_like(ch1wfm)
        self._sample.awg.wfm_send(ch1wfm, marker, marker, self._awg_filepath + self._ch1_filename, self._sample.clock)
        self._sample.awg.wfm_import(self._ch1_filename, self._awg_filepath + self._ch1_filename, 'WFM')
        self._sample.awg.wfm_send(ch2wfm, marker, marker, self._awg_filepath + self._ch2_filename, self._sample.clock)
        self._sample.awg.wfm_import(self._ch2_filename, self._awg_filepath + self._ch2_filename, 'WFM')
        self._sample.awg.set({'ch1_waveform': self._ch1_filename,
                              'ch1_offset': 0,
                              'ch1_amplitude': 2,
                              'ch2_waveform': self._ch2_filename,
                              'ch1_offset': 0,
                              'ch1_amplitude': 2})
        self._sample.awg.run()
        self._sample.awg.set({'ch1_output': 1, 'ch2_output': 1})

    def load_wfm(self, sin_phase=0, update_channels=(True, True), init=False, relamp=1.5, relamp2=1.5):
        if (update_channels[0]):
            t1 = np.linspace(sin_phase, sin_phase + 2 * np.pi, self._sample.clock / self._iq_frequency, endpoint=False)
            ch1wfm = self.ch1(t1)
            marker = np.zeros_like(ch1wfm)
            self._sample.awg.wfm_send(ch1wfm, marker, marker, self._awg_filepath + self._ch1_filename,
                                      self._sample.clock)
            self._sample.awg.wfm_import(self._ch1_filename, self._awg_filepath + self._ch1_filename, 'WFM')
            self._sample.awg.set_ch1_waveform(self._ch1_filename)
        if (update_channels[1]):
            t2 = np.linspace(0, 2 * np.pi, self._sample.clock / self._iq_frequency, endpoint=False)
            ch2wfm = self.ch2(t2)
            marker = np.zeros_like(ch2wfm)
            self._sample.awg.wfm_send(ch2wfm, marker, marker, self._awg_filepath + self._ch2_filename,
                                      self._sample.clock)
            self._sample.awg.wfm_import(self._ch2_filename, self._awg_filepath + self._ch2_filename, 'WFM')
            self._sample.awg.set_ch2_waveform(self._ch2_filename)
        if (init):
            self._sample.awg.set({'ch1_amplitude': relamp, 'ch2_amplitude': relamp2})
            self._sample.awg.run()
            self._sample.awg.set({'ch1_output': 1, 'ch2_output': 1})

    def phaseoptimize(self, offset):
        '''
        phase offset in radians
        '''
        self.load_wfm(sin_phase=offset, update_channels=(True, False))
        self._fsup.sweep()
        return self._fsup.get_marker_level(4)  # assuming that marker4 is at unwanted sideband

    def recalibrate(self, dcx, dcy, x, y, phaseoffset, relamp, relamp2):
        self.cache_valid = False
        if self._swb is not None:  # switching to fsup
            self._swb.set_position('2')
        else:
            logging.warning(
                __name__ + ' : Switch was not switched. Define switchbox instrument via iq.connect_switchbox(swb). If you do not have a switchbox, make sure your cables are connected and go on.')
        self._sample.awg.set_clock(self._sample.clock)
        if not self._FSUP_connected:
            raise ValueError((
                                 'FSUP is possibly not connected. \nIncrease trust_region and maxage to interpolate values or connect FSUP and execute connect_FSUP(fsup)'))
        qkit.flow.start()
        (hx_amp, hx_phase, hy_amp, hy_phase) = (0, 0, 0, 0)
        if self._iq_frequency == 0: logging.warning(
            __name__ + ': Your IQ Frequency is 0. It is better to calibrate with a finite IQ frequency because you will get inconsistent data in the calibration file otherwise. If you calibrate with iq!=0, the right values for iq=0 are extracted.')
        mw_freq = self._f_rounded - self._iq_frequency
        # self._sample.awg.stop()
        self._sample.awg.set({'ch1_output': 0, 'ch2_output': 0, 'runmode': 'CONT'})
        self._sample.qubit_mw_src.set({'frequency': mw_freq, 'power': self._mw_power, 'status': 1})

        print("Recalibrating %s for Frequency: %.2fGHz (MW-Freq: %.2fGHz), MW Power: %.2fdBm" % (
            self.mixer_name, self._sideband_frequency / 1e9, mw_freq / 1e9, self._mw_power))
        sys.stdout.flush()
        frequencies = [mw_freq - 3 * self._iq_frequency, mw_freq - 2 * self._iq_frequency, mw_freq - self._iq_frequency,
                       mw_freq, mw_freq + self._iq_frequency, mw_freq + 2 * self._iq_frequency,
                       mw_freq + 3 * self._iq_frequency]
        self.focus(mw_freq, 1)
        self.load_zeros()
        print("Optimizing DC offsets to reduce leakage when there is no pulse")
        (xold, yold) = (np.inf, np.inf)
        while (np.all(np.around((dcx, dcy), 3) != np.around((xold, yold), 3))):
            (xold, yold) = (dcx, dcy)
            dcx = self.minimize(self.xoptimize, dcx - .002, dcx + .002, .002, 1e-3, final_averages=3, confirmonly=True)[
                0]
            dcy = self.minimize(self.yoptimize, dcy - .002, dcy + .002, .002, 1e-3, final_averages=3, confirmonly=True)[
                0]
        if not self._iq_frequency == 0:
            self.load_wfm(sin_phase=phaseoffset, update_channels=(True, True), relamp=relamp, relamp2=relamp2,
                          init=True)
            optimized = [(self.focus(frequencies[i], 1), self._fsup.get_marker_level(1))[1] for i in
                         range(len(frequencies))]
            optimized_old = [np.inf, np.inf, np.inf, np.inf]
            print("iterating")
            while (optimized_old[2] - optimized[2] > 1 or optimized_old[3] - optimized[3] > 1):
                print ".",
                sys.stdout.flush()
                optimized_old = copy(optimized)
                self.focus(mw_freq, 1)
                (xold, yold) = (np.inf, np.inf)
                while (np.all(np.around((x, y), 3) != np.around((xold, yold), 3))):
                    (xold, yold) = (x, y)
                    x = \
                    self.minimize(self.xoptimize, x - .002, x + .002, .001, 1e-3, final_averages=1, confirmonly=True)[0]
                    y = \
                    self.minimize(self.yoptimize, y - .002, y + .002, .001, 1e-3, final_averages=1, confirmonly=True)[0]
                self.focus(mw_freq - self._iq_frequency, 4)
                phaseoffset = \
                self.minimize(self.phaseoptimize, phaseoffset - .15, phaseoffset + .15, 5e-3, 5e-3, final_averages=1,
                              confirmonly=True)[0]
                relamp = self.minimize(self.relampoptimize, relamp - .01, relamp + .01, .05, 1e-3, final_averages=2,
                                       bounds=(.5, 2), confirmonly=True)[0]
                relamp2 = self.minimize(self.relampoptimize2, relamp2 - .01, relamp2 + .01, .05, 1e-3, final_averages=2,
                                        bounds=(.5, 2), confirmonly=True)[0]
                optimized = [(self.focus(frequencies[i], 1),
                              np.mean([(self._fsup.sweep(), self._fsup.get_marker_level(1))[1] for j in range(5)]))[1]
                             for i in range(len(frequencies))]
        else:
            x, y, phaseoffset, relamp, relamp2 = 0, 0, 0, 1, 1
            optimized = [(self.focus(frequencies[i], 1),
                          np.mean([(self._fsup.sweep(), self._fsup.get_marker_level(1))[1] for j in range(5)]))[1] for i
                         in range(len(frequencies))]
        print("Parameters: DC x: %.1fmV, DC y: %.1fmV AC x: %.1fmV, AC y: %.1fmV phase: %.1fdegree Amplitude: %.3fVpp/%.3fVpp" % (
            dcx * 1e3, dcy * 1e3, x * 1e3, y * 1e3, phaseoffset * 180 / np.pi, relamp, relamp2))
        print(
                "Your Sideband has a power of %.3fdBm, Leakage is %.2fdB lower, other sideband is %.2fdB lower.\nThe largest of the higher harmonics is %.2fdB lower." % (
                    optimized[4], optimized[4] - optimized[3], optimized[4] - optimized[2], np.max((optimized[4] - optimized[0],
                                                                                                    optimized[4] - optimized[1],
                                                                                                    optimized[4] - optimized[4],
                                                                                                    optimized[5] - optimized[6]))))
        data = np.array([np.append(
            (self._f_rounded, mw_freq, self._mw_power, time.time(), dcx, dcy, x, y, phaseoffset, relamp, relamp2),
            optimized)])
        currentdata = copy(data)

        # Make a nice image on the FSUP so that you can see the results of calibration there.
        self._fsup.set({'freqspan': self._iq_frequency * 10,
                        'centerfreq': self._sideband_frequency - self._iq_frequency,
                        'resolutionBW': 5e5,
                        'videoBW': 5e2})
        self._fsup.set_marker(1, self._sideband_frequency - self._iq_frequency)
        self._fsup.set_marker(2, self._sideband_frequency)
        self._fsup.set_marker(3, self._sideband_frequency + self._iq_frequency)
        self._fsup.set_marker(4, self._sideband_frequency + 2 * self._iq_frequency)
        sweeptime = self._fsup.get_sweeptime()
        self._fsup.sweep()
        qkit.flow.sleep(sweeptime)
        if qkit.module_available("matplotlib"):
            plt.plot(self._fsup.get_frequencies() / 1e9, self._fsup.get_trace())
            plt.xlabel('Frequency (GHz)')
            plt.ylabel('Power (dBm)')
            plt.grid()

        if self._swb is not None:  # switching back
            self._swb.set_position('1')
        # reset awg
        [self._sample.awg.set({'ch%i_amplitude' % i: 2, 'ch%i_offset' % i: 0}) for i in (1, 2)]

        try:
            storedvalues = np.loadtxt(os.path.join(qkit.cfg.getget('datadir'), "IQMixer\\%s.cal" % self.mixer_name))
            if np.size(storedvalues) < 30:  # Only one dataset
                storedvalues = [storedvalues]
            # If there was a calibration with the same parameters, remove it
            todelete = []
            for index, t in enumerate(storedvalues):
                if t[0] == self._f_rounded and t[1] == mw_freq and t[2] == self._mw_power:
                    todelete = np.append(todelete, index)
                    print("\nLast time (%s), there have been the following values:" % (time.ctime(t[3])))
                    print("Parameters: DC x: %.1fmV, DC y: %.1fmV phase: %.1fdegree Amplitude: %.3fVpp/%.3fVpp" % (
                        t[6] * 1e3, t[7] * 1e3, t[8] * 180 / np.pi, t[9], t[10]))
                    print(
                            "Your Sideband had a power of %.3fdBm, Leakage was %.2fdB lower, other sideband was %.2fdB lower.\nThe largest of the higher harmonics was %.2fdB lower." % (
                                t[15], t[15] - t[14], t[15] - t[13],
                                np.max((t[15] - t[11], t[15] - t[12], t[15] - t[16], t[15] - t[17]))))

            storedvalues = np.delete(storedvalues, todelete, axis=0)

            data = np.append(storedvalues, data)

        except IOError:
            pass
        data = data.reshape((data.size / 18, 18))
        np.savetxt(os.path.join(qkit.cfg.getget('datadir'), "IQMixer\\%s.cal" % self.mixer_name), data, (
        "%.2f", "%.2f", "%.2f", "%i", "%.3f", "%.3f", "%.3f", "%.3f", "%.6f", "%.3f", "%.3f", "%.4f", "%.4f", "%.4f",
        "%.4f", "%.4f", "%.4f", "%.4f"))
        return currentdata[0]

    def initial_calibrate(self):
        if not self._FSUP_connected:
            raise ValueError((
                                 'FSUP is possibly not connected. \nIncrease trust_region and maxage to interpolate values or connect FSUP and execute connect_FSUP(fsup)'))

        if self._swb is not None:  # switching to fsup
            self._swb.set_position('2')
        else:
            logging.warning(
                __name__ + ' : Switch was not switched. Define switchbox instrument via iq.connect_switchbox(swb). If you do not have a switchbox, make sure your cables are connected and go on.')
        qkit.flow.start()
        (hx_amp, hx_phase, hy_amp, hy_phase, phaseoffset) = (0, 0, 0, 0, 0)
        self._f_rounded = np.round(self._sideband_frequency, -3)
        mw_freq = self._f_rounded - self._iq_frequency
        if self._iq_frequency == 0: logging.warning(
            __name__ + ': Your IQ Frequency is 0. It is better to calibrate with a finite IQ frequency because you will get inconsistent data in the calibration file otherwise. If you calibrate with iq!=0, the right values for iq=0 are extracted.')

        # self._sample.awg.stop()
        self._sample.awg.set({'ch1_output': 0, 'ch2_output': 0, 'runmode': 'CONT'})
        self._sample.qubit_mw_src.set({'frequency': mw_freq, 'power': self._mw_power, 'status': 1})

        print("Starting an initial calibration of %s for Frequency: %.2fGHz (MW-Freq: %.2fGHz), MW Power: %.2fdBm" % (
            self.mixer_name, self._sideband_frequency / 1e9, mw_freq / 1e9, self._mw_power))
        sys.stdout.flush()
        frequencies = [mw_freq - 3 * self._iq_frequency, mw_freq - 2 * self._iq_frequency, mw_freq - self._iq_frequency,
                       mw_freq, mw_freq + self._iq_frequency, mw_freq + 2 * self._iq_frequency,
                       mw_freq + 3 * self._iq_frequency]
        self.focus(mw_freq, 1)
        self.load_zeros()
        (xold, yold) = (np.inf, np.inf)
        print("Finding initial values for DC offsets to reduce leakage when there is no pulse")
        dcx = self.minimize(self.xoptimize, -.02, .02, .01, 5e-3, final_averages=1)[0]
        dcy = self.minimize(self.yoptimize, -.02, .02, .01, 5e-3, final_averages=1)[0]
        while (np.all(np.around((dcx, dcy), 3) != np.around((xold, yold), 3))):
            (xold, yold) = (dcx, dcy)
            dcx = self.minimize(self.xoptimize, dcx - .002, dcx + .002, .002, 1e-3, final_averages=3, confirmonly=True)[
                0]
            dcy = self.minimize(self.yoptimize, dcy - .002, dcy + .002, .002, 1e-3, final_averages=3, confirmonly=True)[
                0]
        if not self._iq_frequency == 0:
            print("Finding initial values for Sine and Cosine waveform parameters")
            self.load_wfm(sin_phase=phaseoffset, update_channels=(True, True), relamp=2, relamp2=2, init=True)
            self.focus(mw_freq, 1)
            (xold, yold) = (0, 0)
            x = self.minimize(self.xoptimize, -.02, .02, .01, 5e-3, final_averages=1)[0]
            y = self.minimize(self.yoptimize, -.02, .02, .01, 5e-3, final_averages=1)[0]
            while (np.all(np.around((x, y), 3) != np.around((xold, yold), 3))):
                (xold, yold) = (x, y)
                x = self.minimize(self.xoptimize, x - .002, x + .002, .002, 2e-3, final_averages=1, verbose=False,
                                  confirmonly=True)[0]
                y = self.minimize(self.yoptimize, y - .002, y + .002, .002, 2e-3, final_averages=1, verbose=False,
                                  confirmonly=True)[0]
            self.focus(mw_freq - self._iq_frequency, 4)
            relamp = self.minimize(self.relampoptimize, 0.2, 2, .3, 10e-3, final_averages=1, bounds=(.5, 2))[0]
            relamp2 = self.minimize(self.relampoptimize2, 0.2, 2, .3, 10e-3, final_averages=1, bounds=(.5, 2))[0]
            phaseoffset = self.minimize(self.phaseoptimize, 0, 1, .2, 5e-3, final_averages=1)[0]
        else:
            x, y, phaseoffset, relamp, relamp2 = 0, 0, 0, 1, 1
        print("Initial parameterset found, starting optimization...")
        return self.recalibrate(dcx, dcy, x, y, phaseoffset, relamp, relamp2)

    def _validate_iq_setting(self):
        '''
        validates iq setting:
        the rounded Tektronix AWG clock devided by the iq frequency needs to be an integer number
        '''

        samples_per_1_iqfreq = float(self._sample.clock) / self._sample.iq_frequency
        if samples_per_1_iqfreq % 2 != 0:  # if number not integer and even
            self._sample.iq_frequency = self._sample.clock / (2 * np.floor(samples_per_1_iqfreq / 2))
            if np.abs(samples_per_1_iqfreq) < 4:
                raise ValueError('samples per iq frequency too small')
            while round(self._sample.iq_frequency, -4) != self._sample.iq_frequency:
                print(self._sample.iq_frequency)
                samples_per_1_iqfreq = float(self._sample.clock) / self._sample.iq_frequency - 1
                if samples_per_1_iqfreq < 4:
                    raise ValueError('samples per iq frequency too small')
                self._sample.iq_frequency = self._sample.clock / (2 * np.floor(samples_per_1_iqfreq / 2))
            logging.warning('Invalid iq frequency for AWG clock setting. Setting iq frequency to ' + str(
                self._sample.iq_frequency * 1e-6) + ' MHz')
        return self._sample.iq_frequency

    def calibrate(self, recalibrate=True, frequency=None, power=None, iq_frequency=None):
        '''
            This function automatically looks up, if this mixer has already been calibrated before:
            - If this frequency was calibrated less than maxage secs ago, it will just load this settings
            - If this frequency was calibrated more than maxage secs ago (or if you set recalibration=True), it will recalibrate, starting at the old parameters
            - If two adjacent frequencies, not further away than interpol_freqspan have been calibrated, interpolate starting values and recalibrate 
            - You can have the same behaviour as above without recalibrating, if you use trust_region instead of interpol_freqspan
            
        '''

        interpol_freqspan = np.max(
            (self.interpol_freqspan, self.trust_region)) / 2  # Half of the span in each direction

        if frequency is None:
            frequency = self._sample.f01
        if power is None:
            power = self._sample.mw_power
        if iq_frequency is None:
            iq_frequency = self._sample.iq_frequency
        self.do_set_sideband_frequency(frequency)
        self.do_set_mw_power(power)
        self.do_set_iq_frequency(iq_frequency)

        self._sample.iq_frequency = self._validate_iq_setting()

        self._f_rounded = np.round(self._sideband_frequency,
                                   -3)  # The FSUP can only resolve 7 digits in frequency, so for Frequencies <10GHz, you can not set frequencies finer than kHz. But as the MW source can, there will be a missmatch if we do not round here.

        try:
            storedvalues = np.loadtxt(os.path.join(qkit.cfg.get('datadir'), "IQMixer\\%s.cal" % self.mixer_name))
            NeedForInterpolation = False
            left_border = -np.inf
            right_border = np.inf
            if len(
                    storedvalues.shape) == 1:  # check dimensionality -> otherwise you will have problems if there is only one line in calibration file.
                storedvalues = [storedvalues]
            for index, t in enumerate(storedvalues):
                if t[0] == self._sideband_frequency and t[1] == self._sideband_frequency - self._iq_frequency and t[
                    2] == self._mw_power:
                    if time.time() - t[3] > self.maxage or recalibrate:  # Calibration is too old
                        print("recalibrating because calibration is too old (%.2f days ago)" % (
                            (time.time() - t[3]) / 24 / 3600))
                        return self.recalibrate(t[4], t[5], t[6], t[7], t[8], t[9], t[10])
                    else:
                        # print "returning known values"
                        return t
                    break
                if np.abs(t[0] - self._sideband_frequency) <= self.interpol_freqspan and t[
                    0] - self._sideband_frequency <= 0:
                    if left_border == -np.inf:
                        left_border = index
                        NeedForInterpolation = True
                    elif storedvalues[left_border][0] < t[0]:
                        left_border = index
                    elif np.abs(storedvalues[left_border][2] - self._mw_power) > np.abs(self._mw_power - t[2]):
                        left_border = index
                if np.abs(t[0] - self._sideband_frequency) <= self.interpol_freqspan and t[
                    0] - self._sideband_frequency >= 0:
                    if right_border == np.inf:
                        right_border = index
                        NeedForInterpolation = True
                    elif storedvalues[right_border][0] > t[0]:
                        right_border = index
                    elif np.abs(storedvalues[right_border][2] - self._mw_power) > np.abs(self._mw_power - t[2]):
                        right_border = index
            if NeedForInterpolation:
                if left_border == -np.inf:  # Within interpol_freqspan, only higher calibrated frequencies could be found
                    print("initial calibration for this frequency required, because only higher frequencies are calibrated yet.")
                    self._f_rounded = np.floor(self._sideband_frequency / 1e4) * 1e4 - 100e6
                    return self.initial_calibrate
                if right_border == +np.inf:  # same for lower frequencies
                    print("initial calibration for this frequency required, because only lower frequencies are calibrated yet.")
                    self._f_rounded = np.ceil(self._sideband_frequency / 1e4) * 1e4 + 100e6
                    return self.initial_calibrate
                if (left_border == right_border):
                    print("Same frequency for different powers found")
                    return self.recalibrate(storedvalues[left_border][4], storedvalues[left_border][5],
                                            storedvalues[left_border][6], storedvalues[left_border][7],
                                            storedvalues[left_border][8], storedvalues[left_border][9],
                                            storedvalues[left_border][10])
                interpolated = storedvalues[left_border] + (storedvalues[right_border] - storedvalues[left_border]) * (
                            self._sideband_frequency - storedvalues[left_border][0]) / (
                                           storedvalues[right_border][0] - storedvalues[left_border][0])
                if (storedvalues[right_border][0] - storedvalues[left_border][
                    0]) <= self.trust_region:  # No need for recalibration
                    if recalibrate:  # Except if we enforce a recalibration
                        return self.recalibrate(interpolated[4], interpolated[5], interpolated[6], interpolated[7],
                                                interpolated[8], interpolated[9], interpolated[10])
                    # print "using interpolated values"
                    return interpolated
                else:
                    print("recalibrate interpolated values")
                    if storedvalues[right_border][0] - self._sideband_frequency > self._sideband_frequency - \
                            storedvalues[left_border][0]:  # sideband_frequency is closer to left border
                        self._f_rounded = np.ceil(self._sideband_frequency / 1e3) * 1e3
                    else:
                        self._f_rounded = np.floor(self._sideband_frequency / 1e3) * 1e3
                    return self.recalibrate(interpolated[4], interpolated[5], interpolated[6], interpolated[7],
                                            interpolated[8], interpolated[9], interpolated[10])
            else:
                print("initial calibration for this frequency required. Think about using the interpol_freqspan option!")
                return self.initial_calibrate()

        except IOError:
            # Mixer has not at all been calibrated before
            # print "initial calibration required"
            # (sideband_frequency,mw_freq,mw_power,time.time(),dcx,dcy,x,y,phaseoffset,relamp,relamp2)
            return (
            self._sample.f01, self._sample.f01 - self._sample.iq_frequency, self._sample.mw_power, 0, 0, 0, 0, 0, np.pi,
            1., 1., -99, -99, -99, -99, -99)
            if self._sideband_frequency != np.round(self._sideband_frequency,
                                                    -3):  # If we only calibrate the rounded frequency, we will get problems later, because there is nothing to interpolate against
                self._f_rounded = np.floor(self._sideband_frequency / 1e3) * 1e3
                dcx, dcy, x, y, phaseoffset, relamp, relamp2 = self.initial_calibrate[4:11]
                self._f_rounded = np.ceil(self._sideband_frequency / 1e3) * 1e3
            return self.recalibrate(dcx, dcy, x, y, phaseoffset, relamp, relamp2)

    def get_calibration(self):
        if self.cache_valid and self.cache_frequency == self._sample.f01 and self.cache_power == self._sample.mw_power and self.cache_iq == self._sample.iq_frequency:
            return self.cache
        else:
            self.cache_frequency = self._sample.f01
            self.cache_power = self._sample.mw_power
            self.cache_iq = self._sample.iq_frequency
            self.cache = self.calibrate(recalibrate=False)
            self.cache_valid = True
            return self.cache

    def convert(self, wfm):
        '''
           wfm is an 1D array of complex values with absolute values <=1,
           where complex phase represents the phase of the later microwave
           Example: With an iq frequency of 100MHz and a samplerate of 6GS/s you have 
           60 Samples per wave, corresponding to 6degree phase resolution
           
        '''
        params = self.get_calibration()

        # content of params:(sideband_frequency,mw_freq,mw_power,time.time(),dcx,dcy,x,y,phaseoffset,relamp,relamp2),optimized
        # self.do_set_sideband_frequency(params[0])
        # self._output_power = params[15]
        # self.do_set_iq_frequency(self._sample.iq_frequency) #This is unnecessary, because it's done in calibrate()
        # self.get_all()
        dcx, dcy, x, y, phaseoffset, relamp, relamp2 = params[4:11]
        if self._iq_frequency == 0: x, y, phaseoffset, relamp, relamp2 = 0., 0., 0., 1., 1.  # homodyne
        t = np.arange(len(wfm)) / self._sample.clock
        # Relamp is Peak-to-Peak
        relamp, relamp2 = relamp / 2, relamp2 / 2
        angle = np.angle(wfm)
        wfm1 = dcx + np.abs(wfm) * (
                    relamp * self.ch1(2 * np.pi * self._iq_frequency * t + angle + phaseoffset) - dcx + x)
        wfm2 = dcy + np.abs(wfm) * (relamp2 * self.ch2(2 * np.pi * self._iq_frequency * t + angle) - dcy + y)
        del dcx, dcy, x, y, phaseoffset, relamp, relamp2, t, angle
        gc.collect()
        return (wfm1, wfm2)
